 /*
  * UAE - The Un*x Amiga Emulator
  *
  * Custom chip emulation
  *
  * Copyright 1995-1998 Bernd Schmidt
  * Copyright 1995 Alessandro Bissacco
  */

 /*
  * Toni Wilen's fixes/updates:
  *
  * - currprefs.chipset type is software controlled ("currprefs.chipset" variable)
  *   set currprefs.chipset before calling custom_init()
  * - BRDRBLNK (Borderblank) in BPLCON3 is emulated
  * - superhires but calcdiw() needs correct FMODE>0 algorithm
  * - experimental PAL/NTSC timing (BEAMCON0 bit 5 and VPOSR chip ID)
  *   currprefs.ntscmode=0, emulates PAL Amiga, currprefs.ntscmode=1, emulates NTSC Amiga (maybe..)
  *   requires following addition to sound code:
  *     void update_sound (void)
  *     {
  *        sample_evtime = (long) maxhpos * maxvpos * vblank_hz / rate;
  *     }     
  * - FMODE BSCAN2 bit is emulated (useless for now, programmed display
  *   timings will be implemented later, I hope...)
  *
  * Working AGA features:
  *  - 0-8 bitplanes
  *  - HAM6, HAM8, EHB, dualplayfield
  *  - bitplane delays 0-63
  *  - lores and hires resolutions
  *  - FMODE and calcdiw() plflinelen calculation
  *  - wide sprites (hacked using multiple normal 16 pixel wide sprites....)
  *  - hires (70ns) sprites
  *  - only non-paletted destination screen modes work. (Anybody want to fix?)
  * Not working:
  *  - sprites outside playfield (sprites are clipped, not easy to implement)
  *  - sprite superhires mode
  *  - AGA superhires playfield is not centered
  * Need fixing:
  *  - sprite code (does not handle well cases when sprite and bitplane resolutions are different)
  */

#include "sysconfig.h"
#include "sysdeps.h"

#include <ctype.h>
#include <assert.h>
#include "config.h"
#include "options.h"
#include "threaddep/penguin.h"
#include "uae.h"
#include "gensound.h"
#include "sounddep/sound.h"
#include "events.h"
#include "memory.h"
#include "custom.h"
#include "readcpu.h"
#include "newcpu.h"
#include "cia.h"
#include "disk.h"
#include "blitter.h"
#include "xwin.h"
#include "joystick.h"
#include "audio.h"
#include "keybuf.h"
#include "serial.h"
#include "osemu.h"
#include "autoconf.h"
#include "gui.h"
#include "picasso96.h"
#include "drawing.h"
#include "savestate.h"

/* SOUND_HAS_PRIORITY is the #define which tells the vsync_handler to disable screen refreshes when sound problems are occurring */
#ifdef SOUND_HAS_PRIORITY
#define CUTOFF 3
static unsigned int n_consecutive_skipped = 0;
static unsigned int total_skipped = 0;
#endif
static int hack_vblank;

#define SPRITE_COLLISIONS

/* Mouse and joystick emulation */

int buttonstate[3];
static int mouse_x, mouse_y;
int joy0button, joy1button;
unsigned int joy0dir, joy1dir;

/* Events */

unsigned long int cycles, nextevent, is_lastline;
static int rpt_did_reset;
struct ev eventtab[ev_max];

frame_time_t vsynctime = 0, vsyncmintime = 0, syncbase = 0;

static int vpos;
static uae_u16 lof;
static int next_lineno;
static enum nln_how nextline_how;
static int lof_changed = 0;

static const int dskdelay = 2; /* FIXME: ??? */

static uae_u32 sprtaba[256],sprtabb[256];

/*
 * Hardware registers of all sorts.
 */

static void custom_wput_1 (int, uaecptr, uae_u32) REGPARAM;

static uae_u16 cregs[256];

uae_u16 intena,intreq;
uae_u16 dmacon;
uae_u16 adkcon; /* used by audio code */

static uae_u32 cop1lc,cop2lc,copcon;
int maxhpos=MAXHPOS_PAL, maxvpos=MAXVPOS_PAL, minfirstline=MINFIRSTLINE_PAL;
int vblank_endline=VBLANK_ENDLINE_PAL, vblank_hz=VBLANK_HZ_PAL;

#define NUM_SPRITES 8
#define MAX_SPRITES 32
static int used_sprites;


/* This is but an educated guess. It seems to be correct, but this stuff
 * isn't documented well. */
enum sprstate { SPR_stop, SPR_restart, SPR_waiting_start, SPR_waiting_stop };
static enum sprstate sprst[NUM_SPRITES];
static int spron[NUM_SPRITES];
static uaecptr sprpt[NUM_SPRITES];
static int sprxpos[MAX_SPRITES], sprvstart[NUM_SPRITES], sprvstop[NUM_SPRITES];

static uae_u32 bpl1dat,bpl2dat,bpl3dat,bpl4dat,bpl5dat,bpl6dat,bpl7dat,bpl8dat;
static uae_s16 bpl1mod,bpl2mod;

static uaecptr bplpt[8];
#ifndef SMART_UPDATE
static char *real_bplpt[8];
#endif

/*static int blitcount[256];  blitter debug */

static struct color_entry current_colors;
static unsigned int bplcon0,bplcon1,bplcon2,bplcon3,bplcon4,beamcon0;
static int nr_planes_from_bplcon0, corrected_nr_planes_from_bplcon0;
static unsigned int diwstrt, diwstop, diwhigh;
static int diwhigh_written;
static unsigned int ddfstrt, ddfstop;
static unsigned int sprdata[MAX_SPRITES], sprdatb[MAX_SPRITES], sprctl[NUM_SPRITES], sprpos[NUM_SPRITES];
static int sprarmed[NUM_SPRITES]; /*sprite_last_drawn_at[MAX_SPRITES];*/
static int real_sprite_width;
static int last_sprite_point, nr_armed;
static uae_u32 dskpt;
static uae_u16 dsklen,dsksync;
static int dsklength;

/* The display and data fetch windows */

enum diw_states
{
    DIW_waiting_start, DIW_waiting_stop
};

static int plffirstline,plflastline,plfstrt,plfstop,plflinelen;
int diwfirstword,diwlastword;
static enum diw_states diwstate, hdiwstate;

/* Sprite collisions */
uae_u16 clxdat, clxcon;
int clx_sprmask;

enum copper_states {
    COP_stop,
    COP_rdelay1, COP_read1, COP_read2,
    COP_bltwait,
    COP_wait1, COP_wait1b, COP_wait2,
};

struct copper {
    /* The current instruction words.  */
    unsigned int i1, i2;
    enum copper_states state;
    /* Instruction pointer.  */
    uaecptr ip;
    int hpos, vpos, count;
    unsigned int ignore_next;
    unsigned int do_move;
    enum diw_states vdiw;
};

static struct copper cop_state;

static void prepare_copper_1 (void);

int dskdmaen; /* used in cia.c */

/*
 * Statistics
 */

/* Used also by bebox.cpp */
unsigned long int msecs = 0, frametime = 0, lastframetime = 0, timeframes = 0, cumulative_time = 0;
static unsigned long int seconds_base, milliseconds_base;
int bogusframe;


static int current_change_set;

#ifdef OS_WITHOUT_MEMORY_MANAGEMENT
/* sam: Those arrays uses around 7Mb of BSS... That seems  */
/* too much for AmigaDOS (uae crashes as soon as one loads */
/* it. So I use a different strategy here (realloc the     */
/* arrays when needed. That strategy might be usefull for  */
/* computer with low memory.                               */
struct sprite_draw  *sprite_positions[2];
struct color_change *color_changes[2];
struct delay_change *delay_changes;
static int max_sprite_draw = 400;
static int delta_sprite_draw = 0;
static int max_color_change = 400;
static int delta_color_change = 0;
static int max_delay_change = 100;
static int delta_delay_change = 0;
#else
struct sprite_draw sprite_positions[2][MAX_REG_CHANGE];
struct color_change color_changes[2][MAX_REG_CHANGE];
/* We don't remember those across frames, that would be too much effort.
 * We simply redraw the line whenever we see one of these. */
struct delay_change delay_changes[MAX_REG_CHANGE];
#endif

struct decision line_decisions[2 * (MAXVPOS+1) + 1];
struct draw_info line_drawinfo[2][2 * (MAXVPOS+1) + 1];
struct color_entry color_tables[2][(MAXVPOS+1) * 2];

struct sprite_draw *curr_sprite_positions, *prev_sprite_positions;
struct color_change *curr_color_changes, *prev_color_changes;
struct draw_info *curr_drawinfo, *prev_drawinfo;
struct color_entry *curr_color_tables, *prev_color_tables;

static int next_color_change, next_sprite_draw, next_delay_change;
static int next_color_entry, remembered_color_entry;
static int color_src_match, color_dest_match, color_compare_result;

/* These few are only needed during/at the end of the scanline, and don't
 * have to be remembered. */
static int decided_bpl1mod, decided_bpl2mod, decided_nr_planes, decided_res;

static char thisline_changed;


#ifdef SMART_UPDATE
#define MARK_LINE_CHANGED do { thisline_changed = 1; } while (0)
#else
#define MARK_LINE_CHANGED do { ; } while (0)
#endif

static struct decision thisline_decision;
static int modulos_added, plane_decided, color_decided, very_broken_program;

/*
 * helper functions
 */

int rpt_available = 0;

void reset_frame_rate_hack (void)
{
    if (! rpt_available) {
        currprefs.m68k_speed = 0;
        return;
    }

    rpt_did_reset = 1;
    is_lastline = 0;
    vsyncmintime = read_processor_time() + vsynctime;
}

static __inline__ void prepare_copper (void)
{
    if (cop_state.vpos > vpos
	|| cop_state.state == COP_stop)
    {
	eventtab[ev_copper].active = 0;
	return;
    }
    prepare_copper_1 ();
}

void check_prefs_changed_custom (void)
{
    currprefs.framerate = changed_prefs.framerate;
    /* Not really the right place... */
    if (currprefs.jport0 != changed_prefs.jport0
	|| currprefs.jport1 != changed_prefs.jport1) {
	currprefs.jport0 = changed_prefs.jport0;
	currprefs.jport1 = changed_prefs.jport1;
	joystick_setting_changed ();
    }
    currprefs.immediate_blits = changed_prefs.immediate_blits;
    currprefs.blits_32bit_enabled = changed_prefs.blits_32bit_enabled;
	
}

static __inline__ void setclr (uae_u16 *p, uae_u16 val)
{
    if (val & 0x8000) {
	*p |= val & 0x7FFF;
    } else {
	*p &= ~val;
    }
}

__inline__ int current_hpos (void)
{
    return cycles - eventtab[ev_hsync].oldcycles;
}

static __inline__ uae_u8 *pfield_xlateptr (uaecptr plpt, int bytecount)
{
    if (!chipmem_bank.check (plpt, bytecount)) {
	static int count = 0;
	if (!count)
	    count++, write_log ("Warning: Bad playfield pointer\n");
	return NULL;
    }
    return chipmem_bank.xlateaddr (plpt);
}

static __inline__ void docols(struct color_entry *colentry)
{
    int i;
    if(!(currprefs.chipset & AGA_CHIPSET)) 
    {
        for (i = 0; i < 32; i++) 
        {
	        int v = colentry->color_regs[i];
	        if (v < 0 || v > 4095) continue;
            colentry->acolors[i] = xcolors[v];
	    }
    }
}

void notice_new_xcolors (void)
{
    int i;

    docols(&current_colors);
/*    docols(&colors_for_drawing);*/
    for (i = 0; i < (MAXVPOS+1)*2; i++) {
	docols(color_tables[0]+i);
	docols(color_tables[1]+i);
    }
}

static void do_sprites (int currvp, int currhp);

static void remember_ctable (void)
{
    if (remembered_color_entry == -1) {
	/* The colors changed since we last recorded a color map. Record a
	 * new one. */
	memcpy (curr_color_tables + next_color_entry, &current_colors, sizeof current_colors);
	remembered_color_entry = next_color_entry++;
    }
    thisline_decision.ctable = remembered_color_entry;
    if (color_src_match == -1 || color_dest_match != remembered_color_entry
	|| line_decisions[next_lineno].ctable != color_src_match)
    {
	/* The remembered comparison didn't help us - need to compare again. */
	int oldctable = line_decisions[next_lineno].ctable;
	int changed = 0;

	if (oldctable == -1) {
	    changed = 1;
	    color_src_match = color_dest_match = -1;
	} else {
	    color_compare_result = memcmp (&prev_color_tables[oldctable].color_regs,
					   &current_colors.color_regs,
					   sizeof current_colors.color_regs) != 0;
	    if (color_compare_result)
		changed = 1;
	    color_src_match = oldctable;
	    color_dest_match = remembered_color_entry;
	}
	thisline_changed |= changed;
    } else {
	/* We know the result of the comparison */
	if (color_compare_result)
	    thisline_changed = 1;
    }
}

static void remember_ctable_for_border (void)
{
    remember_ctable ();
}

/* Called to determine the state of the horizontal display window state
 * machine at the current position. It might have changed since we last
 * checked.  */
static void decide_diw (int hpos)
{
    if (hdiwstate == DIW_waiting_start && thisline_decision.diwfirstword == -1
	&& PIXEL_XPOS (hpos) >= diwfirstword)
    {
	thisline_decision.diwfirstword = diwfirstword;
	hdiwstate = DIW_waiting_stop;
	/* Decide playfield delays only at DIW start, because they don't matter before and
	 * some programs change them after DDF start but before DIW start. */
	thisline_decision.bplcon1 = bplcon1;
	if (thisline_decision.diwfirstword != line_decisions[next_lineno].diwfirstword)
	    MARK_LINE_CHANGED;
	thisline_decision.diwlastword = -1;
    }
    if (hdiwstate == DIW_waiting_stop && thisline_decision.diwlastword == -1
	&& PIXEL_XPOS (hpos) >= diwlastword)
    {
	thisline_decision.diwlastword = diwlastword;
	hdiwstate = DIW_waiting_start;
	if (thisline_decision.diwlastword != line_decisions[next_lineno].diwlastword)
	    MARK_LINE_CHANGED;
    }
}

/* Called when we know that the line is not in the border and we want to draw
 * it.  The data fetch starting position and length are passed as parameters. */
static __inline__ void decide_as_playfield (int startpos, int len)
{
    thisline_decision.which = 1;

    /* The latter condition might be able to happen in interlaced frames. */
    if (vpos >= minfirstline && (thisframe_first_drawn_line == -1 || vpos < thisframe_first_drawn_line))
	thisframe_first_drawn_line = vpos;
    thisframe_last_drawn_line = vpos;

    thisline_decision.plfstrt = startpos;
    thisline_decision.plflinelen = len;

    /* These are for comparison. */
    thisline_decision.bplcon0 = bplcon0;
    thisline_decision.bplcon2 = bplcon2;
    thisline_decision.bplcon3 = bplcon3;
    thisline_decision.bplcon4 = bplcon4;

#ifdef SMART_UPDATE
    if (line_decisions[next_lineno].plfstrt != thisline_decision.plfstrt
	|| line_decisions[next_lineno].plflinelen != thisline_decision.plflinelen
	|| line_decisions[next_lineno].bplcon0 != thisline_decision.bplcon0
	|| line_decisions[next_lineno].bplcon2 != thisline_decision.bplcon2
    || line_decisions[next_lineno].bplcon3 != thisline_decision.bplcon3
    || line_decisions[next_lineno].bplcon4 != thisline_decision.bplcon4
	)
#endif /* SMART_UPDATE */
	thisline_changed = 1;
}

/* Called when we already decided whether the line is playfield or border,
 * but are no longer sure that this was such a good idea. This can make a
 * difference only for very badly written software.
 * 
 * We try to handle two cases:
 * a) bitplane DMA gets turned off during a line which was decided as plane
 * b) bitplane DMA turns on for all or some planes, either after the line was
 *    decided as border, or after we made an incorrect nr_planes decision.
 * 
 * Examples of programs which need this: Majic 12 "Ray of Hope 2" demo
 * (final large scrolltext) and Masterblazer (menu screen).
 * 
 * There's no hope of catching all cases; we'd need a cycle based emulation
 * for that. I'm getting increasingly convinced that that would be a good
 * idea.
 */
static int broken_plane_sub[8];

static void init_broken_program (void)
{
    int i;
    if (very_broken_program)
	return;
    very_broken_program = 1;
    for (i = 0; i < 8; i++)
	broken_plane_sub[i] = i >= decided_nr_planes || thisline_decision.which != 1 ? -1 : 0;
}

static void adjust_broken_program (int hpos)
{
    int tmp1, tmp2;
    int i;
    if (decided_res == RES_HIRES) {
	tmp1 = hpos & 3;
	tmp2 = hpos & ~3;
    } else if(decided_res == RES_SUPERHIRES) {
	tmp1 = hpos & 1;
	tmp2 = hpos & ~1;
    } else {
	tmp1 = hpos & 7;
	tmp2 = hpos & ~7;
    }
    for (i = 0; i < decided_nr_planes; i++) {
	if (broken_plane_sub[i] != -1)
	    continue;
	/* FIXME: superhires, AGA playfields */
	if (decided_res == RES_HIRES) {
	    broken_plane_sub[i] = (tmp2 - thisline_decision.plfstrt) / 4 * 2;
	    switch (i) {
	     case 3: broken_plane_sub[i] += 2; break; /* @@@ break was missing */
	     case 2: broken_plane_sub[i] += (tmp1 >= 3 ? 2 : 0); break;
	     case 1: broken_plane_sub[i] += (tmp1 >= 2 ? 2 : 0); break;
	     case 0: break;
	    }
	} else {
	    broken_plane_sub[i] = (tmp2 - thisline_decision.plfstrt) / 8 * 2;
	    switch (i) {
	     case 5: broken_plane_sub[i] += (tmp1 >= 3 ? 2 : 0); break;
	     case 4: broken_plane_sub[i] += (tmp1 >= 7 ? 2 : 0); break;
	     case 3: broken_plane_sub[i] += (tmp1 >= 2 ? 2 : 0); break;
	     case 2: broken_plane_sub[i] += (tmp1 >= 6 ? 2 : 0); break;
	     case 1: broken_plane_sub[i] += (tmp1 >= 4 ? 2 : 0); break;
	     case 0: break;
	    }
	}
    }
}

static __inline__ void post_decide_line (int hpos)
{
    if (thisline_decision.which == 1
	&& (hpos < maxhpos && hpos < thisline_decision.plfstrt + thisline_decision.plflinelen)
	&& (GET_PLANES(bplcon0) == 0 || !dmaen (DMA_BITPLANE)))
    {
	/* This is getting gross... */
	thisline_decision.plflinelen = hpos - thisline_decision.plfstrt;
	thisline_decision.plflinelen &= 7;
	if (thisline_decision.plflinelen == 0)
	    thisline_decision.which = -1;
	/* ... especially THIS! */
	modulos_added = 1;
	return;
    }

    if (diwstate == DIW_waiting_start
	|| GET_PLANES (bplcon0) == 0
	|| !dmaen (DMA_BITPLANE)
	|| (hpos >= maxhpos || hpos >= thisline_decision.plfstrt + thisline_decision.plflinelen))
	return;

    if (thisline_decision.which == 1
	&& hpos > thisline_decision.plfstrt
	&& decided_nr_planes < nr_planes_from_bplcon0)
    {
	decided_res = ((bplcon0 & 0x8000) == 0x8000) ? RES_HIRES : RES_LORES;
	if(bplcon0 & 0x0040) decided_res = RES_SUPERHIRES;
    	init_broken_program ();
	decided_nr_planes = nr_planes_from_bplcon0;
	thisline_decision.bplcon0 &= 0xFEF;
	thisline_decision.bplcon0 |= bplcon0 & 0xF010;
	adjust_broken_program (hpos);
	MARK_LINE_CHANGED; /* Play safe. */
	return;
    }
    
    if (thisline_decision.which != -1)
	return;

#if 0 /* Can't warn because Kickstart 1.3 does it at reset time ;-) */
    {
	static int warned = 0;
	if (!warned) {
	    write_log ("That program you are running is _really_ broken.\n");
	    warned = 1;
	}
    }
#endif
    init_broken_program ();

    decided_nr_planes = nr_planes_from_bplcon0;
    thisline_decision.bplcon0 &= 0xFEF;
    thisline_decision.bplcon0 |= bplcon0 & 0xF010;

    MARK_LINE_CHANGED; /* Play safe. */
    decide_as_playfield (plfstrt, plflinelen);
    adjust_broken_program (hpos);
}

static void decide_line_1 (int hpos)
{
    do_sprites (vpos, hpos);

    /* Surprisingly enough, this seems to be correct here - putting this into
     * decide_diw() results in garbage. */
    if (diwstate == DIW_waiting_start && vpos == plffirstline) {
	diwstate = DIW_waiting_stop;
    }
    if (diwstate == DIW_waiting_stop && vpos == plflastline) {
	diwstate = DIW_waiting_start;
    }

    if (framecnt != 0) {
/*	thisline_decision.which = -2; This doesn't do anything but hurt, I think. */
	return;
    }

    if (!dmaen(DMA_BITPLANE) || diwstate == DIW_waiting_start || nr_planes_from_bplcon0 == 0) {
	/* We don't want to draw this one. */
	thisline_decision.which = -1;
	thisline_decision.plfstrt = plfstrt;
	thisline_decision.plflinelen = plflinelen;
	return;
    }

    decided_res = (bplcon0 & 0x8000) == 0x8000 ? RES_HIRES : RES_LORES;
    if(bplcon0 & 0x0040) decided_res = RES_SUPERHIRES;
    decided_nr_planes = nr_planes_from_bplcon0;
#if 0
    /* The blitter gets slower if there's high bitplane activity.
     * @@@ but the values must be different. FIXME */
    if (bltstate != BLT_done
	&& ((decided_res && decided_nr_planes > 2)
	    || (!decided_res && decided_nr_planes > 4)))
    {
	int pl = decided_nr_planes;
	unsigned int n = (eventtab[ev_blitter].evtime-cycles);
	if (n > plflinelen)
	    n = plflinelen;
	n >>= 1;
	if (decided_res)
	    pl <<= 1;
	if (pl == 8)
	    eventtab[ev_blitter].evtime += plflinelen;
	else if (pl == 6)
	    eventtab[ev_blitter].evtime += n*2;
	else if (pl == 5)
	    eventtab[ev_blitter].evtime += n*3/2;
	events_schedule();
    }
#endif
    decide_as_playfield (plfstrt, plflinelen);
}

/* Main entry point for deciding how to draw a line. May either do
 * nothing, decide the line as border, or decide the line as playfield. */
static __inline__ void decide_line (int hpos)
{
    if (thisline_decision.which == 0 && hpos >= plfstrt)
	decide_line_1 (hpos);
}

/* Called when a color is about to be changed (write to a color register),
 * but the new color has not been entered into the table yet. */
static void record_color_change (int hpos, int regno, unsigned long value)
{
    /* Early positions don't appear on-screen. */
    if (framecnt != 0 || vpos < minfirstline || hpos < 0x18
	/*|| currprefs.emul_accuracy == 0*/)
	return;

    decide_diw (hpos);
    decide_line (hpos);

    /* See if we can record the color table, but have not done so yet.
     * @@@ There might be a minimal performance loss in case someone changes
     * a color exactly at the start of the DIW. I don't think it can actually
     * fail to work even in this case, but I'm not 100% sure.
     * @@@ There might be a slightly larger performance loss if we're drawing
     * this line as border... especially if there are changes in colors != 0
     * we might end up setting line_changed for no reason. FIXME */
    if (thisline_decision.diwfirstword >= 0
	&& thisline_decision.which != 0)
    {
	if (thisline_decision.ctable == -1)
	    remember_ctable ();
    }

    /* Changes outside the DIW can be ignored if the modified color is not the
     * background color, or if the accuracy is < 2. */
    if ((regno != 0 || currprefs.emul_accuracy < 2)
	&& (diwstate == DIW_waiting_start || thisline_decision.diwfirstword < 0
	    || thisline_decision.diwlastword >= 0
	    || thisline_decision.which == 0
	    || (thisline_decision.which == 1 && (hpos>=maxhpos || hpos >= thisline_decision.plfstrt + thisline_decision.plflinelen))))
	return;

    /* If the border is changed the first time before the DIW, record the
     * original starting border value. */
    if (regno == 0 && thisline_decision.color0 == 0xFFFFFFFFul && thisline_decision.diwfirstword < 0) {
	thisline_decision.color0 = current_colors.color_regs[0];
	if (line_decisions[next_lineno].color0 != value)
	    thisline_changed = 1;
    }
    /* Anything else gets recorded in the color_changes table. */
#ifdef OS_WITHOUT_MEMORY_MANAGEMENT
    if (next_color_change >= max_color_change) {
	++delta_color_change;
	return;
    }
#endif
    curr_color_changes[next_color_change].linepos = hpos;
    curr_color_changes[next_color_change].regno = regno;
    curr_color_changes[next_color_change++].value = value;
}

/* Stupid hack to get some Sanity demos working. */
static void decide_delay (int hpos)
{
    static int warned;

    /* Don't do anything if we're outside the DIW. */
    if (thisline_decision.diwfirstword == -1 || thisline_decision.diwlastword > 0)
	return;
    decide_line (hpos);
    /* Half-hearted attempt to get things right even when post_decide_line() changes
     * the decision afterwards. */
    if (thisline_decision.which != 1) {
	thisline_decision.bplcon1 = bplcon1;
	return;
    }
    /* @@@ Could check here for DDF stopping earlier than DIW */

#ifdef OS_WITHOUT_MEMORY_MANAGEMENT
    if (next_delay_change >= max_delay_change) {
	++delta_delay_change;
	return;
    }
#endif
    delay_changes[next_delay_change].linepos = hpos;
    delay_changes[next_delay_change++].value = bplcon1;
    if (!warned) {
	warned = 1;
	write_log ("Program is torturing BPLCON1.\n");
    }
}

/*
 * The decision which bitplane pointers to use is not made at plfstrt, since
 * data fetch does not start for all planes at this point. Therefore, we wait
 * for the end of the ddf area or the first write to any of the bitplane
 * pointer registers, whichever comes last, before we decide which plane pointers
 * to use.
 * Call decide_line() before this function.
 */
static void decide_plane (int hpos)
{
    int i, bytecount;

    if (framecnt != 0 || plane_decided)
	return;

    if (decided_nr_planes == -1 /* Still undecided */
	|| (hpos < maxhpos && hpos < thisline_decision.plfstrt + thisline_decision.plflinelen))
	return;

    plane_decided = 1;

    bytecount = plflinelen / (RES_SHIFT(decided_res)) * 2;

    if (bytecount > MAX_WORDS_PER_LINE * 2) {
	    /* Can't happen. */
	    static int warned = 0;
	    if (!warned) {
	        write_log ("Mysterious bug in decide_plane(). Please report.\n");
            gui_message ("Mysterious bug in decide_plane(). Please report.\n");
            warned = 1;
        }
	    bytecount = 0;
    }
    if (!very_broken_program) {
	uae_u8 *dataptr = line_data[next_lineno];
	for (i = 0; i < decided_nr_planes; i++, dataptr += MAX_WORDS_PER_LINE*2) {
	    uaecptr pt = bplpt[i];
	    uae_u8 *real_ptr = pfield_xlateptr(pt, bytecount);
	    if (real_ptr == NULL)
		real_ptr = pfield_xlateptr(0, 0);
#ifdef SMART_UPDATE
#if 1
	    if (thisline_changed)
		memcpy (dataptr, real_ptr, bytecount);
	    else
#endif
		thisline_changed |= memcmpy (dataptr, real_ptr, bytecount);
#else
	    real_bplpt[i] = real_ptr;
#endif
	}
    } else {
	uae_u8 *dataptr = line_data[next_lineno];
	for (i = 0; i < decided_nr_planes; i++, dataptr += MAX_WORDS_PER_LINE*2) {
	    uaecptr pt = bplpt[i];
	    uae_u8 *real_ptr;

	    pt -= broken_plane_sub[i];
	    real_ptr = pfield_xlateptr (pt, bytecount);
	    if (real_ptr == NULL)
		real_ptr = pfield_xlateptr(0, 0);
#ifdef SMART_UPDATE
	    if (!thisline_changed)
		thisline_changed |= memcmpy (dataptr, real_ptr, bytecount);
	    else
		memcpy (dataptr, real_ptr, bytecount);
#else
	    real_bplpt[i] = real_ptr;
#endif
	}
    }
}

/*
 * Called from the BPLxMOD routines, after a new value has been written.
 * This routine decides whether the new value is already relevant for the
 * current line.
 */
static void decide_modulos (int hpos)
{
    /* All this effort just for the Sanity WOC demo... */
    decide_line (hpos);
    if (decided_nr_planes != -1 && (hpos >=maxhpos || hpos >= thisline_decision.plfstrt + thisline_decision.plflinelen))
	return;
    decided_bpl1mod = bpl1mod;
    decided_bpl2mod = bpl2mod;
}

/*
 * Add the modulos to the bitplane pointers if data fetch is already
 * finished for the current line.
 * Call decide_plane() before calling this, so that we won't use the
 * new values of the plane pointers for the current line.
 */
static void do_modulos (int hpos)
{
    /* decided_nr_planes is != -1 if this line should be drawn by the
     * display hardware, regardless of whether it fits on the emulated screen.
     */
    if (decided_nr_planes != -1 && plane_decided && !modulos_added
	&& (hpos >= maxhpos || hpos >= thisline_decision.plfstrt + thisline_decision.plflinelen))
    {
	int bytecount = thisline_decision.plflinelen / (RES_SHIFT(decided_res)) * 2;
	int add1, add2, mod;

	/* FMODE BSCAN2-bit emulation */
	if (fmode & 0x4000) {
	    if (((diwstrt >> 8) & 1) ^ (vpos & 1))
		mod = decided_bpl2mod;
		else
		mod = decided_bpl1mod;
	    add1 = add2 = bytecount + mod;
	} else {
	    add1 = bytecount + decided_bpl1mod;
	    add2 = bytecount + decided_bpl2mod;
	}

	if (!very_broken_program) {
	    switch (decided_nr_planes) {
	     case 8: bplpt[7] += add2;
	     case 7: bplpt[6] += add1;
	     case 6: bplpt[5] += add2;
	     case 5: bplpt[4] += add1;
	     case 4: bplpt[3] += add2;
	     case 3: bplpt[2] += add1;
	     case 2: bplpt[1] += add2;
	     case 1: bplpt[0] += add1;
	    }
	} else switch (decided_nr_planes) {
	     case 8: bplpt[7] += add2 - broken_plane_sub[7];
	     case 7: bplpt[6] += add1 - broken_plane_sub[6];
	     case 6: bplpt[5] += add2 - broken_plane_sub[5];
	     case 5: bplpt[4] += add1 - broken_plane_sub[4];
	     case 4: bplpt[3] += add2 - broken_plane_sub[3];
	     case 3: bplpt[2] += add1 - broken_plane_sub[2];
	     case 2: bplpt[1] += add2 - broken_plane_sub[1];
	     case 1: bplpt[0] += add1 - broken_plane_sub[0];
	}

	modulos_added = 1;
    }
}

static __inline__ void record_sprite (int spr, int sprxp)
{
    int pos = next_sprite_draw;
    unsigned int data, datb;

#ifdef OS_WITHOUT_MEMORY_MANAGEMENT
    if(pos >= max_sprite_draw) {
	++delta_sprite_draw;
	return;
    }
#endif
    /* XXX FIXME, this isn't very clever, but it might do */
    for (;;) {
	if (pos == curr_drawinfo[next_lineno].first_sprite_draw)
	    break;
	if (curr_sprite_positions[pos-1].linepos < sprxp)
	    break;
	if (curr_sprite_positions[pos-1].linepos == sprxp
	    && curr_sprite_positions[pos-1].num > spr)
	    break;
	printf("Foo\n");
	pos--;
    }
    if (pos != next_sprite_draw) {
	int pos2 = next_sprite_draw;
	while (pos2 != pos) {
	    curr_sprite_positions[pos2] = curr_sprite_positions[pos2-1];
	    pos2--;
	}
    }
    curr_sprite_positions[pos].linepos = sprxp;
    curr_sprite_positions[pos].num = spr;
    curr_sprite_positions[pos].ctl = sprctl[spr & (NUM_SPRITES-1)];
    data = sprdata[spr];
    datb = sprdatb[spr];
    curr_sprite_positions[pos].datab = ((sprtaba[data & 0xFF] << 16) | sprtaba[data >> 8]
					| (sprtabb[datb & 0xFF] << 16) | sprtabb[datb >> 8]);
    next_sprite_draw++;
}

static void decide_sprites (int hpos)
{
    int nrs[MAX_SPRITES], posns[MAX_SPRITES], count, i;
    int point = PIXEL_XPOS (hpos);

    if (framecnt != 0 || hpos < 0x14 || nr_armed == 0 || point == last_sprite_point)
	return;

    decide_diw (hpos);
    decide_line (hpos);

    if (thisline_decision.which != 1)
	return;

    count = 0;
    for (i = 0; i < used_sprites; i++) {
	int sprxp = sprxpos[i];
	int j, bestp;

	if (!sprarmed[i & (NUM_SPRITES-1)] || sprxp < 0 || sprxp > point || last_sprite_point >= sprxp)
	    continue;
	if ((thisline_decision.diwfirstword >= 0 && sprxp + sprite_width < thisline_decision.diwfirstword)
	    || (thisline_decision.diwlastword >= 0 && sprxp > thisline_decision.diwlastword))
	    continue;

	for (bestp = 0; bestp < count; bestp++) {
	    if (posns[bestp] > sprxp)
		break;
	    if (posns[bestp] == sprxp && nrs[bestp] < i)
		break;
	}
	for (j = count; j > bestp; j--) {
	    posns[j] = posns[j-1];
	    nrs[j] = nrs[j-1];
	}
	posns[j] = sprxp;
	nrs[j] = i;
	count++;
    }
    for (i = 0; i < count; i++)
	record_sprite (nrs[i], posns[i]);
    last_sprite_point = point;
}

/* End of a horizontal scan line. Finish off all decisions that were not
 * made yet. */
static void finish_decisions (void)
{
    struct draw_info *dip;
    struct draw_info *dip_old;
    struct decision *dp;
    int changed;
    int hpos = current_hpos ();

    if (framecnt != 0)
	return;

    decide_diw (hpos);
    if (thisline_decision.which == 0)
	decide_line_1 (hpos);

    /* Large DIWSTOP values can cause the stop position never to be
     * reached, so the state machine always stays in the same state and
     * there's a more-or-less full-screen DIW. */
    if (hdiwstate == DIW_waiting_stop) {
	thisline_decision.diwlastword = max_diwlastword;
	if (thisline_decision.diwlastword != line_decisions[next_lineno].diwlastword)
	    MARK_LINE_CHANGED;
    }

    if (line_decisions[next_lineno].which != thisline_decision.which)
	thisline_changed = 1;
    decide_plane (hpos);

    dip = curr_drawinfo + next_lineno;
    dip_old = prev_drawinfo + next_lineno;
    dp = line_decisions + next_lineno;
    changed = thisline_changed;

    if (thisline_decision.which == 1) {
	record_diw_line (diwfirstword, diwlastword);

	decide_sprites (hpos);

	if (thisline_decision.bplcon1 != line_decisions[next_lineno].bplcon1)
	    changed = 1;
    }

    dip->last_color_change = next_color_change;
    dip->last_delay_change = next_delay_change;
    dip->last_sprite_draw = next_sprite_draw;

    if (thisline_decision.ctable == -1) {
	if (thisline_decision.which == 1)
	    remember_ctable ();
	else
	    remember_ctable_for_border ();
    }
    if (thisline_decision.which == -1 && thisline_decision.color0 == 0xFFFFFFFFul)
	thisline_decision.color0 = current_colors.color_regs[0];

    dip->nr_color_changes = next_color_change - dip->first_color_change;
    dip->nr_sprites = next_sprite_draw - dip->first_sprite_draw;

    if (dip->first_delay_change != dip->last_delay_change)
	changed = 1;
    if (!changed
	&& (dip->nr_color_changes != dip_old->nr_color_changes
	    || (dip->nr_color_changes > 0
		&& memcmp (curr_color_changes + dip->first_color_change,
			   prev_color_changes + dip_old->first_color_change,
			   dip->nr_color_changes * sizeof *curr_color_changes) != 0)))
	changed = 1;
    if (!changed && thisline_decision.which == 1
	&& (dip->nr_sprites != dip_old->nr_sprites
	    || (dip->nr_sprites > 0
		&& memcmp (curr_sprite_positions + dip->first_sprite_draw,
			   prev_sprite_positions + dip_old->first_sprite_draw,
			   dip->nr_sprites * sizeof *curr_sprite_positions) != 0)))
	changed = 1;

    if (changed) {
	thisline_changed = 1;
	*dp = thisline_decision;
    } else
	/* The only one that may differ: */
	dp->ctable = thisline_decision.ctable;
}

/* Set the state of all decisions to "undecided" for a new scanline. */
static void reset_decisions (void)
{
    if (framecnt != 0)
	return;
    thisline_decision.which = 0;
    decided_bpl1mod = bpl1mod;
    decided_bpl2mod = bpl2mod;
    decided_nr_planes = -1;

    /* decided_res shouldn't be touched before it's initialized by decide_line(). */
    thisline_decision.diwfirstword = -1;
    thisline_decision.diwlastword = -2;
    if (hdiwstate == DIW_waiting_stop) {
	thisline_decision.diwfirstword = PIXEL_XPOS (DISPLAY_LEFT_SHIFT/2);
	if (thisline_decision.diwfirstword != line_decisions[next_lineno].diwfirstword)
	    MARK_LINE_CHANGED;
    }
    thisline_decision.ctable = -1;
    thisline_decision.color0 = 0xFFFFFFFFul;

    thisline_changed = 0;
    curr_drawinfo[next_lineno].first_color_change = next_color_change;
    curr_drawinfo[next_lineno].first_delay_change = next_delay_change;
    curr_drawinfo[next_lineno].first_sprite_draw = next_sprite_draw;

    /* memset(sprite_last_drawn_at, 0, sizeof sprite_last_drawn_at); */
    last_sprite_point = 0;
    modulos_added = 0;
    plane_decided = 0;
    color_decided = 0;
    very_broken_program = 0;
}

/* Initialize the decision array, once before the emulator really starts. */
static void init_decisions (void)
{
    size_t i;
    for (i = 0; i < sizeof line_decisions / sizeof *line_decisions; i++) {
	line_decisions[i].which = -2;
    }
}

/* set PAL or NTSC timing variables */
/* vblank_hz can be used to change emulation speed */
/* for example 100 = two times faster than normal A500 speed */
static void init_hz(void)
{
int isntsc;

init_decisions();
isntsc = beamcon0 & 0x20 ? 0 : 1;
if(!isntsc) {
	maxvpos = MAXVPOS_PAL;
	maxhpos = MAXHPOS_PAL;
	minfirstline = MINFIRSTLINE_PAL;
	vblank_endline = VBLANK_ENDLINE_PAL;
	vblank_hz = VBLANK_HZ_PAL;
} else {
	maxvpos = MAXVPOS_NTSC;
	maxhpos = MAXHPOS_NTSC;
	minfirstline = MINFIRSTLINE_NTSC;
	vblank_endline = VBLANK_ENDLINE_NTSC;
	vblank_hz = VBLANK_HZ_NTSC;
}
vsynctime = syncbase / vblank_hz;
update_sound();
write_log("Using %s timing\n",isntsc ? "NTSC" : "PAL");
}

/* Calculate display window and data fetch values from the corresponding
 * hardware registers. */
static void calcdiw (void)
{
    int mask, add, mask2, add2, plfold;
    int hstrt = diwstrt & 0xFF;
    int hstop = diwstop & 0xFF;
    int vstrt = diwstrt >> 8;
    int vstop = diwstop >> 8;

    if (diwhigh_written) {
	hstrt |= ((diwhigh >> 5) & 1) << 8;
	hstop |= ((diwhigh >> 13) & 1) << 8;
	vstrt |= (diwhigh & 7) << 8;
	vstop |= ((diwhigh >> 8) & 7) << 8;
    } else {
	hstop += 0x100;
	if ((vstop & 0x80) == 0)
	    vstop |= 0x100;
    }

    diwfirstword = coord_hw_to_native_x (hstrt - 1);
    diwlastword = coord_hw_to_native_x (hstop - 1);

    if (diwlastword > max_diwlastword)
	diwlastword = max_diwlastword;
    if (diwfirstword < 0)
	diwfirstword = 0;
    if (diwlastword < 0)
	diwlastword = 0;

    plffirstline = vstrt;
    plflastline = vstop;

#if 0
    /* This happens far too often. */
    if (plffirstline < minfirstline) {
	fprintf(stderr, "Warning: Playfield begins before line %d!\n", minfirstline);
	plffirstline = minfirstline;
    }
#endif

#if 0 /* Turrican does this */
    if (plflastline > 313) {
	fprintf (stderr, "Warning: Playfield out of range!\n");
	plflastline = 313;
    }
#endif
    plfstrt = ddfstrt;
    plfstop = ddfstop;
    if (plfstrt < 0x18) plfstrt = 0x18;
    if (plfstop < 0x18) plfstop = 0x18;
    if (plfstop > 0xD8) plfstop = 0xD8;
    if (plfstrt > plfstop) plfstrt = plfstop;

    /* ! If the masking operation is changed, the pfield_doline code could break
     * on some systems (alignment) */
    /* This actually seems to be correct now, at least the non-AGA stuff... */

    switch (fmode & 3)
	{
	case 0:	/* FMODE=0 */
	mask = ~7;
	add = 15;
	mask2 = ~3;
	add2 = 0;
	break;
	case 1:
	case 2: /* FMODE=1/2 */
	if(bplcon0 & 0x0040) {
	    /* no idea.. */
	    mask = ~7;
	    add = 15;
	    mask2 = ~3;
	    add2 = 0;
	} else if (bplcon0 & 0x8000) { 
	    mask = ~15;
	    add = 31;
	    mask2 = ~15;
	    add2 = 4;
	} else {
	    mask = ~15;
	    add = 31;
	    mask2 = ~15;
	    add2 = 8;
	}
	break;
	case 3: /* FMODE=3 */
	if(bplcon0 & 0x0040) {
	    /* not correct */
	    mask = ~7;
	    add = 15;
	    mask2 = ~3;
	    add2 = 0;
	} else if (bplcon0 & 0x8000) {
	    mask = ~15;
	    add = 31;
	    mask2 = ~7;
	    add2 = 4;
	} else {
	    mask = ~31;
	    add = 63;
	    mask2 = ~15;
	    add2 = 8;
	}
	break;
    }
    /* ugh.. This works in every demo and game I have tested, but can't be correct,
     * it looks too complex and stupid, does anybody know how this really works? (TW)
     */
    plfold = plfstrt;
    plfstrt &= mask2;
    plfstrt += add2;
    plfstop += plfstrt - plfold;
    plfstop &= mask2;
    plfstop += add2;
    plflinelen = (plfstop-plfstrt+add) & mask;
//write_log("vpos: %d fmode: %d, hires %d strt1 %d stop1 %d strt2 %d stop2 %d,plflinelen %d\n",vpos,fmode&3,(bplcon0&0x8000)?1:(bplcon0&0x0040)?2:0,ddfstrt,ddfstop,plfstrt,plfstop,plflinelen);
}

/*
 * lores,fmode=3 24 184 = 192
 * hires,fmode=3 40 200 = 176
 * hires,fmode=3 40 216 = 192
 * hires,fmode=1 56 200 = 160
 * hires,fmode=1 56 208 = 160
 * lores,fmode=1 48 200 = 176
 * lores,fmode=1 72 168 = 112 
*/

/* Mousehack stuff */

#define defstepx (1<<16)
#define defstepy (1<<16)
#define defxoffs 0
#define defyoffs 0

static const int docal = 60, xcaloff = 40, ycaloff = 20;
static const int calweight = 3;
static int lastsampledmx, lastsampledmy;
static int lastspr0x,lastspr0y,lastdiffx,lastdiffy,spr0pos,spr0ctl;
static int mstepx,mstepy,xoffs=defxoffs,yoffs=defyoffs;
static int sprvbfl;

int lastmx, lastmy;
int newmousecounters;
int ievent_alive = 0;

static enum { unknown_mouse, normal_mouse, dont_care_mouse, follow_mouse } mousestate;

static void mousehack_setdontcare (void)
{
    if (mousestate == dont_care_mouse)
	return;

    write_log ("Don't care mouse mode set\n");
    mousestate = dont_care_mouse;
    lastspr0x = lastmx; lastspr0y = lastmy;
    mstepx = defstepx; mstepy = defstepy;
}

static void mousehack_setfollow (void)
{
    if (mousestate == follow_mouse)
	return;

    write_log ("Follow sprite mode set\n");
    mousestate = follow_mouse;
    lastdiffx = lastdiffy = 0;
    sprvbfl = 0;
    spr0ctl = spr0pos = 0;
    mstepx = defstepx; mstepy = defstepy;
}

uae_u32 mousehack_helper (void)
{
    int mousexpos, mouseypos;

#ifdef PICASSO96
    if (picasso_on) {
	mousexpos = lastmx - picasso96_state.XOffset;
	mouseypos = lastmy - picasso96_state.YOffset;
    } else
#endif
    {
	if (lastmy >= gfxvidinfo.height)
	    lastmy = gfxvidinfo.height - 1;
	mouseypos = coord_native_to_amiga_y (lastmy) << 1;
	mousexpos = coord_native_to_amiga_x (lastmx);
    }

    switch (m68k_dreg (regs, 0)) {
     case 0:
	return ievent_alive ? -1 : needmousehack ();
     case 1:
	ievent_alive = 10;
	return mousexpos;
     case 2:
	return mouseypos;
    }
    return 0;
}

void togglemouse (void)
{
    switch (mousestate) {
     case dont_care_mouse: mousehack_setfollow (); break;
     case follow_mouse: mousehack_setdontcare (); break;
     default: break; /* Nnnnnghh! */
    }
}

static __inline__ int adjust (int val)
{
    if (val > 127)
	return 127;
    else if (val < -127)
	return -127;
    return val;
}

static void do_mouse_hack (void)
{
    int spr0x = ((spr0pos & 0xff) << 2) | ((spr0ctl & 1) << 1);
    int spr0y = ((spr0pos >> 8) | ((spr0ctl & 4) << 6)) << 1;
    int diffx, diffy;

    if (ievent_alive > 0) {
	mouse_x = mouse_y = 0;
	return;
    }
    switch (mousestate) {
     case normal_mouse:
	diffx = lastmx - lastsampledmx;
	diffy = lastmy - lastsampledmy;
	if (!newmousecounters) {
	    if (diffx > 127) diffx = 127;
	    if (diffx < -127) diffx = -127;
	    mouse_x += diffx;
	    if (diffy > 127) diffy = 127;
	    if (diffy < -127) diffy = -127;
	    mouse_y += diffy;
	}
	lastsampledmx += diffx; lastsampledmy += diffy;
	break;

     case dont_care_mouse:
	diffx = adjust (((lastmx - lastspr0x) * mstepx) >> 16);
	diffy = adjust (((lastmy - lastspr0y) * mstepy) >> 16);
	lastspr0x = lastmx; lastspr0y = lastmy;
	mouse_x += diffx; mouse_y += diffy;
	break;

     case follow_mouse:
	if (sprvbfl && sprvbfl-- > 1) {
	    int mousexpos, mouseypos;

	    if ((lastdiffx > docal || lastdiffx < -docal)
		&& lastspr0x != spr0x
		&& spr0x > plfstrt*4 + 34 + xcaloff
		&& spr0x < plfstop*4 - xcaloff)
	    {
		int val = (lastdiffx << 16) / (spr0x - lastspr0x);
		if (val >= 0x8000)
		    mstepx = (mstepx * (calweight - 1) + val) / calweight;
	    }
	    if ((lastdiffy > docal || lastdiffy < -docal)
		&& lastspr0y != spr0y
		&& spr0y > plffirstline + ycaloff
		&& spr0y < plflastline - ycaloff)
	    {
		int val = (lastdiffy << 16) / (spr0y - lastspr0y);
		if (val >= 0x8000)
		    mstepy = (mstepy * (calweight - 1) + val) / calweight;
	    }
	    if (lastmy >= gfxvidinfo.height)
		lastmy = gfxvidinfo.height-1;
	    mouseypos = coord_native_to_amiga_y (lastmy) << 1;
	    mousexpos = coord_native_to_amiga_x (lastmx);
	    diffx = adjust ((((mousexpos + xoffs - spr0x) & ~1) * mstepx) >> 16);
	    diffy = adjust ((((mouseypos + yoffs - spr0y) & ~1) * mstepy) >> 16);
	    lastspr0x = spr0x; lastspr0y = spr0y;
	    lastdiffx = diffx; lastdiffy = diffy;
	    mouse_x += diffx; mouse_y += diffy;
	}
	break;
	
     default:
	abort ();
    }
}

 /*
  * register functions
  */

static __inline__ uae_u16 DMACONR (void)
{
    return (dmacon | (bltstate==BLT_done ? 0 : 0x4000)
	    | (blt_info.blitzero ? 0x2000 : 0));
}
#define DMACON_R DMACONR

static __inline__ uae_u16 INTENAR (void)
{
    return intena;
}
#define INTENA_R INTENAR

uae_u16 INTREQR (void)
{
    return intreq | (currprefs.use_serial ? 0x0001 : 0);
}
#define INTREQ_R INTREQR

static __inline__ uae_u16 ADKCONR (void)
{
    return adkcon;
}
#define ADKCON_R ADKCONR

static __inline__ uae_u16 VPOSR (void)
{
    if(currprefs.chipset & AGA_CHIPSET) return (vpos >> 8) | lof | (currprefs.ntscmode? 0x3300 : 0x2300);
    if(currprefs.chipset & ECS_AGNUS) return (vpos >> 8) | lof | (currprefs.ntscmode ? 0x3000: 0x2000);
    return (vpos >> 8) | lof | (currprefs.ntscmode ? 0x1000 : 0x0000);
}
static void VPOSW (uae_u16 v)
{
    if (lof != (v & 0x8000))
    {
	    lof_changed = 1;
        lof = v & 0x8000;
    }
    if((v&1)&&vpos>0) hack_vblank=vpos;
}
static uae_u16 VPOSW_R (void) { return lof; }
#define VPOSW_W VPOSW

static __inline__ uae_u16 VHPOSR (void)
{
    return (vpos << 8) | current_hpos();
}

static uae_u16 LISAID (void)
{
    if(currprefs.chipset & AGA_CHIPSET) return 0xF8;
    if(currprefs.chipset & ECS_DENISE) return 0xFC;
    return 0xffff;
}

static __inline__ void COP1LCH (uae_u16 v) { cop1lc = (cop1lc & 0xffff) | ((uae_u32)v << 16); }
static __inline__ void COP1LCL (uae_u16 v) { cop1lc = (cop1lc & ~0xffff) | (v & 0xfffe); }
static __inline__ void COP2LCH (uae_u16 v) { cop2lc = (cop2lc & 0xffff) | ((uae_u32)v << 16); }
static __inline__ void COP2LCL (uae_u16 v) { cop2lc = (cop2lc & ~0xffff) | (v & 0xfffe); }
static uae_u16 COP1LCH_R (void) { return (uae_u16)(cop1lc>>16); }
static uae_u16 COP1LCL_R (void) { return (uae_u16)(cop1lc&0xffff); }
static uae_u16 COP2LCH_R (void) { return (uae_u16)(cop2lc>>16); }
static uae_u16 COP2LCL_R (void) { return (uae_u16)(cop2lc&0xffff); }
#define COP1LCH_W COP1LCH
#define COP1LCL_W COP1LCL
#define COP2LCH_W COP2LCH
#define COP2LCL_W COP2LCL

static void COPJMP1 (uae_u16 a)
{
    cop_state.ip = cop1lc;
    cop_state.do_move = 0;
    cop_state.ignore_next = 0;
    cop_state.state = COP_read1;
    cop_state.vpos = vpos;
    cop_state.hpos = current_hpos () & ~1;
    cop_state.count = current_hpos () & ~1;
    prepare_copper ();
    if (eventtab[ev_copper].evtime == cycles && eventtab[ev_copper].active)
	abort ();
    events_schedule ();
}

static void COPJMP2 (uae_u16 a)
{
    cop_state.ip = cop2lc;
    cop_state.do_move = 0;
    cop_state.ignore_next = 0;
    cop_state.state = COP_read1;
    cop_state.vpos = vpos;
    cop_state.hpos = current_hpos () & ~1;
    cop_state.count = current_hpos () & ~1;
    prepare_copper ();
    if (eventtab[ev_copper].evtime == cycles && eventtab[ev_copper].active)
	abort ();
    events_schedule ();
}

static __inline__ void COPCON (uae_u16 a)
{
    copcon = a;
}
static uae_u16 COPCON_R (void) { return copcon; }
#define COPCON_W COPCON

static void DMACON (uae_u16 v)
{
    int i, need_resched = 0;

    uae_u16 oldcon = dmacon;

    decide_line (current_hpos ());
    setclr(&dmacon,v); dmacon &= 0x1FFF;
    /* ??? post_decide_line (); */

    /* FIXME? Maybe we need to think a bit more about the master DMA enable
     * bit in these cases. */
    if ((dmacon & DMA_COPPER) > (oldcon & DMA_COPPER)) {
	cop_state.ip = cop1lc;
	cop_state.do_move = 0;
	cop_state.ignore_next = 0;
	cop_state.state = COP_read1;
	cop_state.vpos = vpos;
	cop_state.hpos = current_hpos () & ~1;
	cop_state.count = current_hpos () & ~1;
	prepare_copper ();
	if (eventtab[ev_copper].evtime == cycles && eventtab[ev_copper].active)
	    abort ();
	need_resched = 1;
    }
    if ((dmacon & DMA_BLITPRI) > (oldcon & DMA_BLITPRI) && bltstate != BLT_done) {
	static int count = 0;
	if (!count) {
	    count = 1;
	    write_log ("warning: program is doing blitpri hacks.\n");
	}
	regs.spcflags |= SPCFLAG_BLTNASTY;
    }

    update_audio ();

    for (i = 0; i < 4; i++) {
	struct audio_channel_data *cdp = audio_channel + i;

	cdp->dmaen = (dmacon & 0x200) && (dmacon & (1<<i));
	if (cdp->dmaen) {
	    if (cdp->state == 0) {
		cdp->state = 1;
		cdp->pt = cdp->lc;
		cdp->wper = cdp->per;
		cdp->wlen = cdp->len;
		cdp->data_written = 2;
		cdp->evtime = eventtab[ev_hsync].evtime - cycles;
	    }
	} else {
	    if (cdp->state == 1 || cdp->state == 5) {
		cdp->state = 0;
		cdp->current_sample = 0;
	    }
	}
    }

    if (need_resched)
	events_schedule();
}

static void DMACON_W (uae_u16 v)
{
dmacon=v&~(0x2000|0x4000);
blt_info.blitzero=(v&0x2000)?1:0;
}

/*static int trace_intena = 0;*/

static __inline__ void INTENA (uae_u16 v)
{
/*    if (trace_intena)
	fprintf (stderr, "INTENA: %04x\n", v);*/
    setclr(&intena,v); regs.spcflags |= SPCFLAG_INT;
}
static void INTENA_W (uae_u16 v) { intena = v; }

void INTREQ (uae_u16 v)
{
    setclr(&intreq,v);
    regs.spcflags |= SPCFLAG_INT;
    if (( v & 0x8800) == 0x0800)
	serdat &= 0xbfff;
}
static void INTREQ_W (uae_u16 v) { intreq = v; }

static void ADKCON (uae_u16 v)
{
    unsigned long t;

    update_audio ();

    setclr (&adkcon,v);
    t = adkcon | (adkcon >> 4);
    audio_channel[0].adk_mask = (((t >> 0) & 1) - 1);
    audio_channel[1].adk_mask = (((t >> 1) & 1) - 1);
    audio_channel[2].adk_mask = (((t >> 2) & 1) - 1);
    audio_channel[3].adk_mask = (((t >> 3) & 1) - 1);
}

static void ADKCON_W (uae_u16 v) { adkcon=v; }

static void BEAMCON0 (uae_u16 v)
{
    v &= 0x20;
    if (beamcon0 == v) return;
    beamcon0 = v;
    if (beamcon0 & (0x4000|0x1000|0x0080))
	write_log("Program is trying to modify display timing parameters\n");
    init_hz();
}
static uae_u16 BEAMCON0_R (void) { return beamcon0; }
#define BEAMCON0_W BEAMCON0

static void FMODE (uae_u16 v)
{
    if((currprefs.chipset & AGA_CHIPSET) && fmode!=v) {
	fmode = v;
	switch((fmode>>2)&3)
	    {
	    case 0:
	    real_sprite_width=16>>3;
	    used_sprites=8;
	    break;
	    case 1:
	    case 2:
	    real_sprite_width=32>>3;
	    used_sprites=16;
	    break;
	    case 3:
	    real_sprite_width=64>>3;
	    used_sprites=32;
	    break;
	}
	calcdiw ();
    }
}
static uae_u16 FMODE_R (void) { return fmode; }
#define FMODE_W FMODE

static void BPLPTH (int hpos, uae_u16 v, int num)
{
    decide_line (hpos);
    decide_plane (hpos);
    do_modulos (hpos);
    bplpt[num] = (bplpt[num] & 0xffff) | ((uae_u32)v << 16);
}
static uae_u16 BPLPTH_R (int num) { return (uae_u16)(bplpt[num]>>16); }
static void BPLPTH_W (int num, uae_u16 v) { bplpt[num] = (bplpt[num] & 0xffff) | ((uae_u32)v << 16); }

static void BPLPTL (int hpos, uae_u16 v, int num)
{
    decide_line (hpos);
    decide_plane (hpos);
    do_modulos (hpos);
    bplpt[num] = (bplpt[num] & ~0xffff) | (v & ~1);
}
static uae_u16 BPLPTL_R (int num) { return (uae_u16)(bplpt[num]&0xffff); }
static void BPLPTL_W (int num, uae_u16 v) { bplpt[num] = (bplpt[num] & ~0xffff) | (v & ~1); }

static void check_bblank(void)
{
int oldbb = borderblank;
borderblank = 0;
if((bplcon3 & 32) && (bplcon0 & 1) && (currprefs.chipset & (ECS_AGNUS|AGA_CHIPSET))) borderblank=1;
if(borderblank != oldbb) init_decisions(); // Is this correct? (TW)
}

static void BPLCON0 (int hpos, uae_u16 v)
{
    int ov;

    if(!(currprefs.chipset & AGA_CHIPSET)) {
 	v &= 0xFF4F;
 	/* The Sanity WOC demo needs this at one place (at the end of the "Party Effect")
 	* Disable bitplane DMA if someone tries to do more than 4 Hires bitplanes. */
 	if ((v & 0xF000) > 0xC000) v &= 0xFFF;
 	/* Don't want 7 lores planes either. */
 	if ((v & 0x8000) == 0 && (v & 0x7000) == 0x7000) v &= 0xEFFF;
	/* no superhires or ecs_ena without ECS DENISE or AGA currprefs.chipset */
	if (!(currprefs.chipset & ECS_DENISE)) bplcon0 &= ~0x0041;
    }

    if (bplcon0 == v)
	return;
    decide_line (hpos);
    nr_planes_from_bplcon0 = GET_PLANES (v);
    ov = bplcon0 & 0x8000;
    bplcon0 = v;
    /* plflinelen depends on hires/lores  */
    if((v & 0x8000) != ov) calcdiw();
    check_bblank();
    if(currprefs.chipset & AGA_CHIPSET) {
 	/* It's not clear how the copper timings are affected by the number
 	* of bitplanes on AGA machines */
 	corrected_nr_planes_from_bplcon0 = 4;
    } else {
 	corrected_nr_planes_from_bplcon0 = nr_planes_from_bplcon0 << (bplcon0 & 0x8000 ? 1 : bplcon0 & 0x0040 ? 2 : 0);
    }
    post_decide_line (hpos);
}
static uae_u16 BPLCON0_R (void) { return bplcon0; }
static void BPLCON0_W (uae_u16 v) { bplcon0=v; }

static __inline__ void BPLCON1 (int hpos, uae_u16 v)
{
    if (bplcon1 == v)
	return;
    decide_diw (hpos);
    bplcon1 = v;
    decide_delay (hpos);
}
static uae_u16 BPLCON1_R (void) { return bplcon1; }
static void BPLCON1_W (uae_u16 v) { bplcon1=v; }

static __inline__ void BPLCON2 (int hpos, uae_u16 v)
{
    if (bplcon2 == v)
    	return;
    decide_line (hpos);
    bplcon2 = v;
}
static uae_u16 BPLCON2_R (void) { return bplcon2; }
static void BPLCON2_W (uae_u16 v) { bplcon2=v; }

static __inline__ void BPLCON3 (int hpos, uae_u16 v)
{
    if (bplcon3 == v)
    	return;
    decide_line (hpos);
    bplcon3 = v;
    check_bblank();
}
static uae_u16 BPLCON3_R (void) { return bplcon3; }
static void BPLCON3_W (uae_u16 v) { bplcon3=v; }

static __inline__ void BPLCON4 (int hpos, uae_u16 v)
{
    if (bplcon4 == v)
    	return;
    decide_line (hpos);
    bplcon4 = v;
}
static uae_u16 BPLCON4_R (void) { return bplcon4; }
static void BPLCON4_W (uae_u16 v) { bplcon4=v; }

static void BPL1MOD (int hpos, uae_u16 v)
{
    v &= ~1;
    if ((uae_s16)bpl1mod == (uae_s16)v)
	return;
    bpl1mod = v;
    decide_modulos (hpos);
}
static uae_u16 BPL1MOD_R (void) { return bpl1mod; }
static void BPL1MOD_W (uae_u16 v) { bpl1mod=v; }

static void BPL2MOD (int hpos, uae_u16 v)
{
    v &= ~1;
    if ((uae_s16)bpl2mod == (uae_s16)v)
	return;
    bpl2mod = v;
    decide_modulos (hpos);
}
static uae_u16 BPL2MOD_R (void) { return bpl2mod; }
static void BPL2MOD_W (uae_u16 v) { bpl2mod=v; }

/* We could do as well without those... */
static __inline__ void BPL1DAT (uae_u16 v) { bpl1dat = v; }
static __inline__ void BPL2DAT (uae_u16 v) { bpl2dat = v; }
static __inline__ void BPL3DAT (uae_u16 v) { bpl3dat = v; }
static __inline__ void BPL4DAT (uae_u16 v) { bpl4dat = v; }
static __inline__ void BPL5DAT (uae_u16 v) { bpl5dat = v; }
static __inline__ void BPL6DAT (uae_u16 v) { bpl6dat = v; }
static __inline__ void BPL7DAT (uae_u16 v) { bpl7dat = v; }
static __inline__ void BPL8DAT (uae_u16 v) { bpl8dat = v; }

static void DIWSTRT (int hpos, uae_u16 v)
{
    if (diwstrt == v && ! diwhigh_written)
	return;
    decide_line (hpos);
    diwhigh_written = 0;
    diwstrt = v;
    calcdiw ();
}
static uae_u16 DIWSTRT_R (void) { return diwstrt; }
static void DIWSTRT_W (uae_u16 v) { diwstrt=v; }

static void DIWSTOP (int hpos, uae_u16 v)
{
    if (diwstop == v && ! diwhigh_written)
	return;
    decide_line (hpos);
    diwhigh_written = 0;
    diwstop = v;
    calcdiw ();
}
static uae_u16 DIWSTOP_R (void) { return diwstop; }
static void DIWSTOP_W (uae_u16 v) { diwstop=v; }

static void DIWHIGH (int hpos, uae_u16 v)
{
    if (diwhigh_written && diwhigh == v)
	return;
    decide_line (hpos);
    diwhigh_written = 1;
    diwhigh = v;
    calcdiw ();
}
static uae_u16 DIWHIGH_R (void) { return diwhigh; }
static void DIWHIGH_W (uae_u16 v) { diwhigh=v; }

static void DDFSTRT (int hpos, uae_u16 v)
{
    v &= (currprefs.chipset & AGA_CHIPSET) ? 0x1FF : 0xFF;
    if (ddfstrt == v)
	return;
    decide_line (hpos);
    ddfstrt = v;
    calcdiw ();
}
static uae_u16 DDFSTRT_R (void) { return ddfstrt; }
static void DDFSTRT_W (uae_u16 v) { ddfstrt=v; }

static void DDFSTOP (int hpos, uae_u16 v)
{
    v &= (currprefs.chipset & AGA_CHIPSET) ? 0x1FF : 0xFF;
    if (ddfstop == v)
	return;
    decide_line (hpos);
    ddfstop = v;
    calcdiw ();
}
static uae_u16 DDFSTOP_R (void) { return ddfstop; }
static void DDFSTOP_W (uae_u16 v) { ddfstop=v; }

static void BLTADAT (uae_u16 v)
{
    maybe_blit();

    blt_info.bltadat = v;
}
static uae_u16 BLTADAT_R (void) { return blt_info.bltadat; }
static void BLTADAT_W (uae_u16 v) { blt_info.bltadat=v; }

/*
 * "Loading data shifts it immediately" says the HRM. Well, that may
 * be true for BLTBDAT, but not for BLTADAT - it appears the A data must be
 * loaded for every word so that AFWM and ALWM can be applied.
 */
static void BLTBDAT (uae_u16 v)
{
    maybe_blit();

    if (bltcon1 & 2)
	blt_info.bltbhold = v << (bltcon1 >> 12);
    else
	blt_info.bltbhold = v >> (bltcon1 >> 12);
    blt_info.bltbdat = v;
}
static uae_u16 BLTBDAT_R (void) { return blt_info.bltbdat; }
static void BLTBDAT_W (uae_u16 v)
{
    if (bltcon1 & 2)
	blt_info.bltbhold = v << (bltcon1 >> 12);
    else
	blt_info.bltbhold = v >> (bltcon1 >> 12);
    blt_info.bltbdat=v;
}
static void BLTCDAT (uae_u16 v) { maybe_blit (); blt_info.bltcdat = v; }
static uae_u16 BLTCDAT_R (void) { return blt_info.bltcdat; }
static void BLTCDAT_W (uae_u16 v) { blt_info.bltcdat=v; }

static void BLTAMOD (uae_u16 v) { maybe_blit (); blt_info.bltamod = (uae_s16)(v & 0xFFFE); }
static void BLTBMOD (uae_u16 v) { maybe_blit (); blt_info.bltbmod = (uae_s16)(v & 0xFFFE); }
static void BLTCMOD (uae_u16 v) { maybe_blit (); blt_info.bltcmod = (uae_s16)(v & 0xFFFE); }
static void BLTDMOD (uae_u16 v) { maybe_blit (); blt_info.bltdmod = (uae_s16)(v & 0xFFFE); }
static uae_u16 BLTAMOD_R (void) { return blt_info.bltamod; }
static uae_u16 BLTBMOD_R (void) { return blt_info.bltbmod; }
static uae_u16 BLTCMOD_R (void) { return blt_info.bltcmod; }
static uae_u16 BLTDMOD_R (void) { return blt_info.bltdmod; }
static void BLTAMOD_W (uae_u16 v) { blt_info.bltamod=v; }
static void BLTBMOD_W (uae_u16 v) { blt_info.bltbmod=v; }
static void BLTCMOD_W (uae_u16 v) { blt_info.bltcmod=v; }
static void BLTDMOD_W (uae_u16 v) { blt_info.bltdmod=v; }

static void BLTCON0 (uae_u16 v) { maybe_blit (); bltcon0 = v; blinea_shift = v >> 12; }
static uae_u16 BLTCON0_R (void) { return bltcon0; }
static void BLTCON0_W (uae_u16 v) { blinea_shift = v >>12; }
/* The next category is "Most useless hardware register".
 * And the winner is... */
static void BLTCON0L (uae_u16 v) { maybe_blit (); bltcon0 = (bltcon0 & 0xFF00) | (v & 0xFF); }
static uae_u16 BLTCON0L_R (void) { return bltcon0 & 0xff; }
static void BLTCON0L_W (uae_u16 v) {}

static void BLTCON1 (uae_u16 v) { maybe_blit (); bltcon1 = v; }
static uae_u16 BLTCON1_R (void) { return bltcon1; }
static void BLTCON1_W (uae_u16 v) { bltcon1=v; }

static void BLTAFWM (uae_u16 v) { maybe_blit (); blt_info.bltafwm = v; }
static void BLTALWM (uae_u16 v) { maybe_blit (); blt_info.bltalwm = v; }
static uae_u16 BLTAFWM_R (void) { return blt_info.bltafwm; }
static uae_u16 BLTALWM_R (void) { return blt_info.bltalwm; }
static void BLTAFWM_W (uae_u16 v) { blt_info.bltafwm = v; }
static void BLTALWM_W (uae_u16 v) { blt_info.bltalwm = v; }

static void BLTAPTH (uae_u16 v) { maybe_blit (); bltapt = (bltapt & 0xffff) | ((uae_u32)v << 16); }
static void BLTAPTL (uae_u16 v) { maybe_blit (); bltapt = (bltapt & ~0xffff) | (v & 0xFFFE); }
static void BLTBPTH (uae_u16 v) { maybe_blit (); bltbpt = (bltbpt & 0xffff) | ((uae_u32)v << 16); }
static void BLTBPTL (uae_u16 v) { maybe_blit (); bltbpt = (bltbpt & ~0xffff) | (v & 0xFFFE); }
static void BLTCPTH (uae_u16 v) { maybe_blit (); bltcpt = (bltcpt & 0xffff) | ((uae_u32)v << 16); }
static void BLTCPTL (uae_u16 v) { maybe_blit (); bltcpt = (bltcpt & ~0xffff) | (v & 0xFFFE); }
static void BLTDPTH (uae_u16 v) { maybe_blit (); bltdpt = (bltdpt & 0xffff) | ((uae_u32)v << 16); }
static void BLTDPTL (uae_u16 v) { maybe_blit (); bltdpt = (bltdpt & ~0xffff) | (v & 0xFFFE); }
static uae_u16 BLTAPTH_R (void) { return (uae_u16)(bltapt>>16); }
static uae_u16 BLTAPTL_R (void) { return (uae_u16)(bltapt&0xffff); }
static uae_u16 BLTBPTH_R (void) { return (uae_u16)(bltbpt>>16); }
static uae_u16 BLTBPTL_R (void) { return (uae_u16)(bltbpt&0xffff); }
static uae_u16 BLTCPTH_R (void) { return (uae_u16)(bltcpt>>16); }
static uae_u16 BLTCPTL_R (void) { return (uae_u16)(bltcpt&0xffff); }
static uae_u16 BLTDPTH_R (void) { return (uae_u16)(bltdpt>>16); }
static uae_u16 BLTDPTL_R (void) { return (uae_u16)(bltdpt&0xffff); }
static void BLTAPTH_W (uae_u16 v) { bltapt = (bltapt & 0xffff) | ((uae_u32)v << 16); }
static void BLTAPTL_W (uae_u16 v) { bltapt = (bltapt & ~0xffff) | (v & 0xFFFE); }
static void BLTBPTH_W (uae_u16 v) { bltbpt = (bltbpt & 0xffff) | ((uae_u32)v << 16); }
static void BLTBPTL_W (uae_u16 v) { bltbpt = (bltbpt & ~0xffff) | (v & 0xFFFE); }
static void BLTCPTH_W (uae_u16 v) { bltcpt = (bltcpt & 0xffff) | ((uae_u32)v << 16); }
static void BLTCPTL_W (uae_u16 v) { bltcpt = (bltcpt & ~0xffff) | (v & 0xFFFE); }
static void BLTDPTH_W (uae_u16 v) { bltdpt = (bltdpt & 0xffff) | ((uae_u32)v << 16); }
static void BLTDPTL_W (uae_u16 v) { bltdpt = (bltdpt & ~0xffff) | (v & 0xFFFE); }

static void BLTSIZE (uae_u16 v)
{
    bltsize = v;

    maybe_blit ();

    blt_info.vblitsize = bltsize >> 6;
    blt_info.hblitsize = bltsize & 0x3F;
    if (!blt_info.vblitsize) blt_info.vblitsize = 1024;
    if (!blt_info.hblitsize) blt_info.hblitsize = 64;

    bltstate = BLT_init;
    do_blitter ();
}
static uae_u16 BLTSIZE_R (void) { return bltsize; }
static void BLTSIZE_W (uae_u16 v) {}

static void BLTSIZV (uae_u16 v)
{
    maybe_blit ();
    oldvblts = v & 0x7FFF;
}
static uae_u16 BLTSIZV_R (void) { return blt_info.vblitsize; }
static void BLTSIZV_W (uae_u16 v) {}

static void BLTSIZH (uae_u16 v)
{
    maybe_blit ();
    blt_info.hblitsize = v & 0x7FF;
    blt_info.vblitsize = oldvblts;
    if (!blt_info.vblitsize) blt_info.vblitsize = 32768;
    if (!blt_info.hblitsize) blt_info.hblitsize = 0x800;
    bltstate = BLT_init;
    do_blitter ();
}
static uae_u16 BLTSIZH_R (void) { return blt_info.hblitsize; }
static void BLTSIZH_W (uae_u16 v) {}

static void widesprite(int sprxp,int num)
{
    sprxpos[3*NUM_SPRITES+num]=coord_hw_to_native_x(sprxp+3*16);
    sprxpos[2*NUM_SPRITES+num]=coord_hw_to_native_x(sprxp+2*16);
    sprxpos[1*NUM_SPRITES+num]=coord_hw_to_native_x(sprxp+1*16);
} 

static __inline__ void SPRxCTL_1 (uae_u16 v, int num)
{
    int sprxp, sprxp2;
    sprctl[num] = v;
    nr_armed -= sprarmed[num];
    sprarmed[num] = 0;
    if (sprpos[num] == 0 && v == 0) {
	sprst[num] = SPR_stop;
	spron[num] = 0;
    } else
	sprst[num] = SPR_waiting_start;

    sprxp2= (sprpos[num] & 0xFF) * 2 + (v & 1);
    sprxp = coord_hw_to_native_x (sprxp2);
    sprxpos[num] = sprxp;
    /* AGA wide sprite hack */
    if(currprefs.chipset & AGA_CHIPSET) widesprite(sprxp2,num);
    sprvstart[num] = (sprpos[num] >> 8) | ((sprctl[num] << 6) & 0x100);
    sprvstop[num] = (sprctl[num] >> 8) | ((sprctl[num] << 7) & 0x100);
}
static uae_u16 SPRxCTL_R (int num) { return sprctl[num]; }
static void SPRxCTL_W (int num, uae_u16 v) {}

static __inline__ void SPRxPOS_1 (uae_u16 v, int num)
{
    int sprxp, sprxp2;
    sprpos[num] = v;
    sprxp2 = (v & 0xFF) * 2 + (sprctl[num] & 1);
    sprxp = coord_hw_to_native_x (sprxp2);
    sprxpos[num] = sprxp;
    sprvstart[num] = (sprpos[num] >> 8) | ((sprctl[num] << 6) & 0x100);
    /* AGA wide sprite hack */
    if(currprefs.chipset & AGA_CHIPSET) widesprite(sprxp2,num);
 }
static uae_u16 SPRxPOS_R (int num) { return sprpos[num]; }
static void SPRxPOS_W (int num, uae_u16 v) {}

static __inline__ void SPRxDATA_1 (uae_u16 v, int num)
{
    sprdata[num] = v;
    nr_armed += 1 - sprarmed[num];
    sprarmed[num] = 1;
}

static __inline__ void SPRxDATA_2 (uae_u16 v, int num)
{
    sprdata[num] = v;
}
static uae_u16 SPRxDATA_R (int num) { return sprdata[num]; }
#define SPRxDATA_W SPRxDATA_1

static __inline__ void SPRxDATB_1 (uae_u16 v, int num)
{
    sprdatb[num] = v;
}
static __inline__ void SPRxDATB_2 (uae_u16 v, int num)
{
    sprdatb[num] = v;
}
static uae_u16 SPRxDATB_R (int num) { return sprdatb[num]; }
#define SPRxDATB_W SPRxDATB_1

static void SPRxDATA (int hpos, uae_u16 v, int num) { decide_sprites (hpos); SPRxDATA_1 (v, num); }
static void SPRxDATB (int hpos, uae_u16 v, int num) { decide_sprites (hpos); SPRxDATB_1 (v, num); }
static void SPRxCTL (int hpos, uae_u16 v, int num) { decide_sprites (hpos); SPRxCTL_1 (v, num); }
static void SPRxPOS (int hpos, uae_u16 v, int num) { decide_sprites (hpos); SPRxPOS_1 (v, num); }
static void SPRxPTH (int hpos, uae_u16 v, int num)
{
    decide_sprites (hpos);
    sprpt[num] &= 0xffff;
    sprpt[num] |= (uae_u32)v << 16;

    if (sprst[num] == SPR_stop || vpos < vblank_endline)
	sprst[num] = SPR_restart;
    spron[num] = 1;
}
static uae_u16 SPRxPTH_R (int num) { return (uae_u16)(sprpt[num]>>16); }
static void SPRxPTH_W (int num, uae_u16 v) {}

static void SPRxPTL (int hpos, uae_u16 v, int num)
{
    decide_sprites (hpos);
    sprpt[num] &= ~0xffff;
    sprpt[num] |= v;

    if (sprst[num] == SPR_stop || vpos < vblank_endline)
	sprst[num] = SPR_restart;
    spron[num] = 1;
}
static uae_u16 SPRxPTL_R (int num) { return (uae_u16)(sprpt[num]&0xffff); }
static void SPRxPTL_W (int num, uae_u16 v) {}

static void CLXCON (uae_u16 v)
{
    clxcon = v;
    clx_sprmask = (((v >> 15) << 7) | ((v >> 14) << 5) | ((v >> 13) << 3) | ((v >> 12) << 1) | 0x55);
}
static uae_u16 CLXCON_R (void) { return clxcon; }
#define CLXCON_W CLXCON
static uae_u16 CLXCON2_R (void) { return 0; }
static void CLXCON2_W (uae_u16 v) {}

static uae_u16 CLXDAT (void)
{
    uae_u16 v = clxdat;
    clxdat = 0;
    return v;
}
static uae_u16 CLXDAT_R (void) { return clxdat; }
static void CLXDAT_W (uae_u16 v) { clxdat=v; }
static uae_u16 COLOR_READ (int num)
{
   int cr, cg, cb, colreg;
   uae_u16 cval;
   if (!(currprefs.chipset & AGA_CHIPSET) || !(bplcon2 & 0x0100)) return(0xffff);
   colreg = ((bplcon3 >> 13) & 7) * 32 + num;
   cr = current_colors.color_regs[colreg] >> 16;
   cg = (current_colors.color_regs[colreg] >> 8) & 0xFF;
   cb = current_colors.color_regs[colreg] & 0xFF;
   if (bplcon3 & 0x200)
	cval = ((cr & 15) << 12) | ((cg & 15) << 4) | ((cb & 15) << 0);
    else
	cval = ((cr >> 4) << 12) | ((cg >> 4) << 4) | ((cb >> 4) << 0);
   return cval;
}

static void COLOR_WRITE (int hpos, uae_u16 v, int num)
{
   v &= 0xFFF;
   if(currprefs.chipset & AGA_CHIPSET) {

	int r,g,b;
	int cr,cg,cb;
	int colreg;
	uae_u32 cval;

	/* is writing disabled when RDRAM=1 ? */
	if (bplcon2 & 0x0100) return;

	colreg = ((bplcon3 >> 13) & 7) * 32 + num;
	r = (v & 0xF00) >> 8;
	g = (v & 0xF0) >> 4;
	b = (v & 0xF) >> 0;
	cr = current_colors.color_regs[colreg] >> 16;
	cg = (current_colors.color_regs[colreg] >> 8) & 0xFF;
	cb = current_colors.color_regs[colreg] & 0xFF;

	if (bplcon3 & 0x200) {
	    cr &= 0xF0; cr |= r;
	    cg &= 0xF0; cg |= g;
	    cb &= 0xF0; cb |= b;
	} else {
	    cr = r + (r << 4);
	    cg = g + (g << 4);
	    cb = b + (b << 4);
	}
	cval = (cr << 16) | (cg << 8) | cb;
	if (cval == current_colors.color_regs[colreg])
	    return;

	/* Call this with the old table still intact. */
	record_color_change (hpos, colreg, cval);
	remembered_color_entry = -1;
	current_colors.color_regs[colreg] = cval;
   } else {
    if (current_colors.color_regs[num] == v)
	    return;
	/* Call this with the old table still intact. */
	record_color_change (hpos, num, v);
	remembered_color_entry = -1;
	current_colors.color_regs[num] = v;
	current_colors.acolors[num] = xcolors[v];
    }
}
static uae_u32 COLOR_R (int num) { return current_colors.color_regs[num]; }
static void COLOR_W (int num, uae_u32 v) { current_colors.color_regs[num]=v; }

static void DSKSYNC (uae_u16 v) { dsksync = v; }
static uae_u16 DSKSYNC_R (void) { return dsksync; }
#define DSKSYNC_W DSKSYNC
static void DSKDAT (uae_u16 v) { write_log ("DSKDAT written. Not good.\n"); }
static uae_u16 DSKDAT_R (void) { return 0; }
static void DSKPTH (uae_u16 v) { dskpt = (dskpt & 0xffff) | ((uae_u32)v << 16); }
static void DSKPTL (uae_u16 v) { dskpt = (dskpt & ~0xffff) | (v); }
static uae_u16 DSKPTH_R (void) { return (uae_u16)(dskpt>>16); }
static uae_u16 DSKPTL_R (void) { return (uae_u16)(dskpt&0xffff); }
#define DSKPTH_W DSKPTH
#define DSKPTL_W DSKPTL

static void DSKLEN (uae_u16 v)
{
    if (v & 0x8000) {
	dskdmaen = dskdmaen == 1 ? 2 : 1;
    } else {
	dskdmaen = 0;
	if (eventtab[ev_diskblk].active)
	    write_log ("warning: Disk DMA aborted!\n");
	eventtab[ev_diskblk].active = 0;
	events_schedule();

    }
    dsklen = dsklength = v; dsklength &= 0x3fff;
    if (dskdmaen == 2 && dsksync != 0x4489 && (adkcon & 0x400)) {
	write_log ("Non-standard sync: %04x len: %x\n", dsksync, dsklength);
    }
    if (dskdmaen > 1) {
	if (dsklen & 0x4000) {
	    eventtab[ev_diskblk].active = 1;
	    eventtab[ev_diskblk].oldcycles = cycles;
	    eventtab[ev_diskblk].evtime = 40 + cycles; /* ??? */
	    events_schedule();
	} else {
	    int result = DISK_PrepareReadMFM(dsklength, dsksync, adkcon & 0x400);
	    if (result) {
		eventtab[ev_diskblk].active = 1;
		eventtab[ev_diskblk].oldcycles = cycles;
		eventtab[ev_diskblk].evtime = result + cycles;
		events_schedule();
	    }
	}
    }
}
static uae_u16 DSKLEN_R (void) { return dsklen; }
static void DSKLEN_W (uae_u16 v) {}

static uae_u16 DSKBYTR (void)
{
    uae_u16 v = (dsklen >> 1) & 0x6000;
    uae_u16 mfm, byte;
    if (DISK_GetData (&mfm, &byte))
	v |= 0x8000;
    v |= byte;
    if (dsksync == mfm)
	v |= 0x1000;
    return v;
}

static uae_u16 DSKDATR (void)
{
    uae_u16 mfm, byte;
    DISK_GetData (&mfm, &byte);
    return mfm;
}

static uae_u16 potgo_value;

static void POTGO (uae_u16 v)
{
    potgo_value = v;
}
static uae_u16 POTGO_R (void) { return potgo_value; }
#define POTGO_W POTGO

static uae_u16 POTGOR (void)
{
    uae_u16 v = (potgo_value | (potgo_value << 1)) & 0xAA00;
    v |= v >> 1;

    if (JSEM_ISMOUSE (0, &currprefs)) {
	if (buttonstate[2])
	    v &= 0xFBFF;

	if (buttonstate[1])
	    v &= 0xFEFF;
    } else if (JSEM_ISJOY0 (0, &currprefs) || JSEM_ISJOY1 (0, &currprefs)) {
	if (joy0button & 2) v &= 0xfbff;
	if (joy0button & 4) v &= 0xfeff;
    }

    if (JSEM_ISJOY0 (1, &currprefs) || JSEM_ISJOY1 (1, &currprefs)) {
	if (joy1button & 2) v &= 0xbfff;
	if (joy1button & 4) v &= 0xefff;
    }

    return v;
}

static uae_u16 POT0DAT (void)
{
    static uae_u16 cnt = 0;
    if (JSEM_ISMOUSE (0, &currprefs)) {
	if (buttonstate[2])
	    cnt = ((cnt + 1) & 0xFF) | (cnt & 0xFF00);
	if (buttonstate[1])
	    cnt += 0x100;
    }

    return cnt;
}
static uae_u16 JOY0DAT (void)
{
    if (JSEM_ISMOUSE (0, &currprefs)) {
	do_mouse_hack ();
	return ((uae_u8)mouse_x) + ((uae_u16)mouse_y << 8);
    }
    return joy0dir;
}
static uae_u16 JOY1DAT (void)
{
    if (JSEM_ISMOUSE (1, &currprefs)) {
	do_mouse_hack ();
	return ((uae_u8)mouse_x) + ((uae_u16)mouse_y << 8);
    }
    return joy1dir;
}
static void JOYTEST (uae_u16 v)
{
    if (JSEM_ISMOUSE (0, &currprefs)) {
	mouse_x = v & 0xFC;
	mouse_y = (v >> 8) & 0xFC;
    }
}
static uae_u16 JOYTEST_R (void) { return 0; }

/*
 * Here starts the copper code. Can you believe it used to be worse?
 */

static __inline__ void copper_adjust_diw (struct copper *cst)
{
    if (cst->vdiw == DIW_waiting_start && vpos == plffirstline)
	cst->vdiw = DIW_waiting_stop;
    if (cst->vdiw == DIW_waiting_stop && vpos == plflastline)
	cst->vdiw = DIW_waiting_start;
}

/* Determine which cycles are available for the copper in a display
 * with a agiven number of planes.  */
static int cycles_for_plane[9][8] = {
    { 0, -1, 0, -1, 0, -1, 0, -1 },
    { 0, -1, 0, -1, 0, -1, 0, -1 },
    { 0, -1, 0, -1, 0, -1, 0, -1 },
    { 0, -1, 0, -1, 0, -1, 0, -1 },
    { 0, -1, 0, -1, 0, -1, 0, -1 },
    { 0, -1, 1, -1, 0, -1, 0, -1 },
    { 0, -1, 1, -1, 0, -1, 1, -1 },
    { 1, -1, 1, -1, 1, -1, 1, -1 },
    { 1, -1, 1, -1, 1, -1, 1, -1 }
};

static unsigned int waitmasktab[256];

#define COP_OFFSET 4

static __inline__ int copper_in_playfield (enum diw_states diw, int hpos)
{
    hpos -= COP_OFFSET;
    return diw == DIW_waiting_stop && hpos >= plfstrt && hpos < plfstrt + plflinelen;
}

static __inline__ int copper_cant_read (enum diw_states diw, int hpos, int planes)
{
    int t;

    if (hpos >= ((maxhpos - 2) & ~1))
	return 1;

    if(currprefs.chipset & AGA_CHIPSET) return 0; /* FIXME */

    if (! copper_in_playfield (diw, hpos))
	return 0;

    t = cycles_for_plane[planes][hpos & 7];
#if 0
    if (t == -1)
	abort ();
#endif
    return t;
}

static void update_copper_1 (int until_hpos)
{
    unsigned int vp, hp, vcmp, hcmp;
    int c_hpos = cop_state.hpos;
    int c_vpos = cop_state.vpos;

    vp = vpos & (((cop_state.i2 >> 8) & 0x7F) | 0x80);
    hp = cop_state.count & (cop_state.i2 & 0xFE);
    vcmp = ((cop_state.i1 & (cop_state.i2 | 0x8000)) >> 8);
    hcmp = (cop_state.i1 & cop_state.i2 & 0xFE);

    if (cop_state.state == COP_wait2)
	cop_state.state = COP_wait1b;

    for (;;) {
	if (c_hpos == (maxhpos & ~1)) {
	    c_hpos = 0;
	    cop_state.count = 0;
	    c_vpos++;
	}

	if (! dmaen (DMA_COPPER))
	    cop_state.state = COP_stop;

	if (c_vpos > vpos
	    || c_hpos > until_hpos
	    || cop_state.state == COP_stop)
	    break;

	if (c_hpos - COP_OFFSET == plfstrt)
	    copper_adjust_diw (&cop_state);

	switch (cop_state.state) {
	 case COP_rdelay1:
	    cop_state.state = COP_read1;
	    break;

	 case COP_read1:
	    if (copper_cant_read (cop_state.vdiw, c_hpos, corrected_nr_planes_from_bplcon0))
		break;

	    if (cop_state.do_move) {
		cop_state.do_move = 0;
 		if (cop_state.i1 < (copcon & 2 ? ((currprefs.chipset & AGA_CHIPSET) ? 0 : 0x40u) : 0x80u)) {
		    cop_state.state = COP_stop;
		    break;
		}
		switch (cop_state.i1) {
		 case 0x088:
		    cop_state.ip = cop1lc;
		    cop_state.state = COP_rdelay1;
		    break;
		 case 0x08A:
		    cop_state.ip = cop2lc;
		    cop_state.state = COP_rdelay1;
		    break;
		 default:
		    custom_wput_1 (c_hpos, cop_state.i1, cop_state.i2);
		    break;
		}
		if (cop_state.state != COP_read1)
		    break;
	    }
	    cop_state.i1 = chipmem_bank.wget (cop_state.ip);
	    cop_state.ip += 2;
	    cop_state.state = COP_read2;
	    break;

	 case COP_read2:
	    if (copper_cant_read (cop_state.vdiw, c_hpos, corrected_nr_planes_from_bplcon0))
		break;
	    cop_state.i2 = chipmem_bank.wget (cop_state.ip);
	    cop_state.ip += 2;
	    if (cop_state.ignore_next) {
		cop_state.ignore_next = 0;
		cop_state.state = COP_read1;
		break;
	    }
	    if ((cop_state.i1 & 1) == 0) {
		cop_state.state = COP_read1;
		cop_state.do_move = 1;
	    } else {
		vp = vpos & (((cop_state.i2 >> 8) & 0x7F) | 0x80);
		hp = cop_state.count & (cop_state.i2 & 0xFE);
		vcmp = ((cop_state.i1 & (cop_state.i2 | 0x8000)) >> 8);
		hcmp = (cop_state.i1 & cop_state.i2 & 0xFE);

		if ((cop_state.i2 & 1) == 0) {
		    cop_state.state = COP_wait1;
		    if (cop_state.i1 == 0xFFFF && cop_state.i2 == 0xFFFE)
			cop_state.state = COP_stop;
		} else {
		    /* Skip instruction.  */
		    if ((vp > vcmp || (vp == vcmp && hp >= hcmp))
			&& ((cop_state.i2 & 0x8000) != 0 || ! (DMACONR() & 0x4000)))
			cop_state.ignore_next = 1;
		    cop_state.state = COP_read1;
		}
	    }
	    break;

	 case COP_wait1:
	    if (copper_cant_read (cop_state.vdiw, c_hpos, corrected_nr_planes_from_bplcon0))
		break;

	    /* fall through */

	 case COP_wait1b:
	    {
		int do_wait2 = cop_state.state == COP_wait1b;
		cop_state.state = COP_wait2;
		if (vp == vcmp && corrected_nr_planes_from_bplcon0 < 8) {
		    int t = cop_state.count + 2;
		    int next_count = (t & waitmasktab[cop_state.i2 & 0xfe]) | hcmp;
		    int nexthpos;
		    if (next_count < t)
			next_count = t;
		    nexthpos = c_hpos + next_count - cop_state.count;
		    if (nexthpos < (maxhpos & ~1)) {
			if (c_hpos - COP_OFFSET < plfstrt && nexthpos - COP_OFFSET >= plfstrt)
			    copper_adjust_diw (&cop_state);
			c_hpos = nexthpos;
			cop_state.count = next_count;
			do_wait2 = 1;
		    }
		}
		if (! do_wait2)
		    break;
	    }

	    /* fall through */

	 case COP_wait2:
	    if (vp < vcmp) {
		if (c_hpos - COP_OFFSET < plfstrt)
		    copper_adjust_diw (&cop_state);

		c_vpos++;
		c_hpos = 0;
		cop_state.count = 0;
		continue;
	    }
	    hp = cop_state.count & (cop_state.i2 & 0xFE);
	    if (vp == vcmp && hp < hcmp)
		break;
	    /* Now we know that the comparisons were successful.  */
	    if ((cop_state.i2 & 0x8000) == 0x8000 || ! (DMACONR() & 0x4000))
		cop_state.state = COP_read1;
	    else {
		/* We need to wait for the blitter.  It won't stop while
		 * we're in update_copper, so we _could_ as well proceed to
		 * until_hpos in one big step.  There are some tricky
		 * issues to be considered, though, so use the slow method
		 * for now.  */
		cop_state.state = COP_bltwait;
	    }
	    break;

	 default:
	    /* Delay cycles.  */
	    break;	    
	}

	c_hpos += 2;
	if (corrected_nr_planes_from_bplcon0 < 8 || ! copper_in_playfield (cop_state.vdiw, c_hpos))
	    cop_state.count += 2;
    }
    cop_state.hpos = c_hpos;
    cop_state.vpos = c_vpos;
}

static __inline__ void update_copper (int until_hpos)
{
    if (cop_state.vpos > vpos
	|| cop_state.hpos > until_hpos
	|| cop_state.state == COP_stop)
	return;
    update_copper_1 (until_hpos);
}

static int dangerous_reg (int reg)
{
    /* Safe:
     * Bitplane pointers, control registers, modulos and data.
     * Sprite pointers, control registers, and data.
     * Color registers.  */
    if (reg >= 0xE0 && reg < 0x1C0)
	return 0;
    return 1;
}

static void prepare_copper_1 (void)
{
    struct copper cst = cop_state;
    unsigned int vp, hp, vcmp, hcmp;

    vp = vpos & (((cop_state.i2 >> 8) & 0x7F) | 0x80);
    hp = cop_state.count & (cop_state.i2 & 0xFE);
    vcmp = ((cop_state.i1 & (cop_state.i2 | 0x8000)) >> 8);
    hcmp = (cop_state.i1 & cop_state.i2 & 0xFE);

    if (cst.state == COP_wait2)
	cst.state = COP_wait1b;

    for (;;) {
	if (cst.hpos == (maxhpos & ~1)) {
	    cst.hpos = 0;
	    cst.count = 0;
	    cst.vpos++;
	}

	if (! dmaen (DMA_COPPER))
	    cst.state = COP_stop;

	if (cst.vpos > vpos
	    || cst.state == COP_stop)
	{
	    eventtab[ev_copper].active = 0;
	    return;
	}

	if (cst.hpos - COP_OFFSET == plfstrt)
	    copper_adjust_diw (&cst);

	switch (cst.state) {
	 case COP_rdelay1:
	    cst.state = COP_read1;
	    break;

	 case COP_read1:
	    if (copper_cant_read (cst.vdiw, cst.hpos, corrected_nr_planes_from_bplcon0))
		break;

	    if (cst.do_move) {
		cst.do_move = 0;
 		if (cst.i1 < (copcon & 2 ? ((currprefs.chipset & AGA_CHIPSET) ? 0 : 0x40u) : 0x80u)) {
		    cst.state = COP_stop;
		    eventtab[ev_copper].active = 0;
		    return;
		} else  if (dangerous_reg (cst.i1)) {
		    eventtab[ev_copper].active = 1;
		    eventtab[ev_copper].oldcycles = cycles;
		    eventtab[ev_copper].evtime = cycles + cst.hpos - current_hpos () + (cst.vpos - cop_state.vpos) * maxhpos;
		    return;
		}
	    }

	    cst.i1 = chipmem_bank.wget (cst.ip);
	    cst.ip += 2;
	    cst.state = COP_read2;
	    break;

	 case COP_read2:
	    if (copper_cant_read (cst.vdiw, cst.hpos, corrected_nr_planes_from_bplcon0))
		break;
	    cst.i2 = chipmem_bank.wget (cst.ip);
	    cst.ip += 2;
	    if (cst.ignore_next) {
		cst.ignore_next = 0;
		cst.state = COP_read1;
		break;
	    }
	    if ((cst.i1 & 1) == 0) {
		cst.state = COP_read1;
		cst.do_move = 1;
	    } else {
		vp = vpos & (((cst.i2 >> 8) & 0x7F) | 0x80);
		hp = cst.count & (cst.i2 & 0xFE);
		vcmp = ((cst.i1 & (cst.i2 | 0x8000)) >> 8);
		hcmp = (cst.i1 & cst.i2 & 0xFE);

		if ((cst.i2 & 1) == 0) {
		    cst.state = COP_wait1;
		    if (cst.i1 == 0xFFFF && cst.i2 == 0xFFFE)
			cst.state = COP_stop;
		} else {
		    /* Skip instruction.  */
		    if ((vp > vcmp || (vp == vcmp && hp >= hcmp))
			&& ((cst.i2 & 0x8000) != 0 || ! (DMACONR() & 0x4000)))
			cst.ignore_next = 1;
		    cst.state = COP_read1;
		}
	    }
	    break;

	 case COP_wait1:
	    if (copper_cant_read (cst.vdiw, cst.hpos, corrected_nr_planes_from_bplcon0))
		break;

	    /* fall through */

	 case COP_wait1b:
	    {
		int do_wait2 = cst.state == COP_wait1b;
		cst.state = COP_wait2;
		if (vp == vcmp && corrected_nr_planes_from_bplcon0 < 8) {
		    int t = cst.count + 2;
		    int next_count = (t & waitmasktab[cst.i2 & 0xfe]) | hcmp;
		    int nexthpos;
		    if (next_count < t)
			next_count = t;
		    nexthpos = cst.hpos + next_count - cst.count;
		    if (nexthpos < (maxhpos & ~1)) {
			if (cst.hpos - COP_OFFSET < plfstrt && nexthpos - COP_OFFSET >= plfstrt)
			    copper_adjust_diw (&cst);
			cst.hpos = nexthpos;
			cst.count = next_count;
			do_wait2 = 1;
		    }
		}
		if (! do_wait2)
		    break;
	    }

	    /* fall through */
	    
	 case COP_wait2:
	    if (vp < vcmp) {
		if (cst.hpos - COP_OFFSET < plfstrt)
		    copper_adjust_diw (&cst);

		cst.vpos++;
		cst.count = 0;
		cst.hpos = 0;
		continue;
	    }
	    hp = cst.count & (cst.i2 & 0xFE);
	    if (vp == vcmp && hp < hcmp)
		break;
#if 0
	    if (nexthpos != -1)
		if (nexthpos != cst.count)
		    printf ("ERROR\n");
	    nexthpos = -1;
#endif
	    /* Now we know that the comparisons were successful.  */
	    if ((cst.i2 & 0x8000) == 0x8000 || ! (DMACONR() & 0x4000))
		cst.state = COP_read1;
	    else {
		/* We need to wait for the blitter.  It won't stop while
		 * we're in update_copper, so we _could_ as well proceed to
		 * until_hpos in one big step.  There are some tricky
		 * issues to be considered, though, so use the slow method
		 * for now.  */
		cst.state = COP_bltwait;
	    }
	    break;

	 default:
	    /* Delay cycles.  */
	    break;	    
	}

	cst.hpos += 2;
	if (corrected_nr_planes_from_bplcon0 < 8 || ! copper_in_playfield (cst.vdiw, cst.hpos))
	    cst.count += 2;
    }
}

void blitter_done_notify (void)
{
    if (cop_state.state != COP_bltwait)
	return;
    update_copper (current_hpos ());
    cop_state.state = COP_wait2;
    eventtab[ev_copper].active = 1;
    eventtab[ev_copper].oldcycles = cycles;
    eventtab[ev_copper].evtime = (cycles + 2) & ~1;
    events_schedule ();
}

static void do_copper (void)
{
    int hpos = current_hpos ();
    update_copper (hpos);
    prepare_copper ();
    if (eventtab[ev_copper].evtime == cycles && eventtab[ev_copper].active)
	abort ();
}

static void diskblk_handler (void)
{
    regs.spcflags |= SPCFLAG_DISK;
    eventtab[ev_diskblk].active = 0;
}

void do_disk (void)
{
    if (dskdmaen != 2 && (regs.spcflags & SPCFLAG_DISK)) {
	static int warned = 0;
	if (!warned)
	    warned++, write_log ("BUG!\n");
	return;
    }
    if (dmaen(0x10)){
	if (dsklen & 0x4000) {
	    if (!chipmem_bank.check (dskpt, 2*dsklength)) {
		write_log ("warning: Bad disk write DMA pointer\n");
	    } else {
		uae_u8 *mfmp = get_real_address (dskpt);
		int i;
		DISK_InitWrite();

		for (i = 0; i < dsklength; i++) {
		    uae_u16 d = (*mfmp << 8) + *(mfmp+1);
		    mfmwrite[i] = d;
		    mfmp += 2;
		}
		DISK_WriteData(dsklength);
	    }
	} else {
	    DISK_ReadMFM (dskpt);
	}
	regs.spcflags &= ~SPCFLAG_DISK;
	INTREQ(0x9002);
	dskdmaen = -1;
    }
}

static void do_sprites (int currvp, int currhp)
{
    int i;
    int maxspr;

#if 0
    if (currvp == 0)
	return;
#else
    /* I don't know whether this is right. Some programs write the sprite pointers
     * directly at the start of the copper list. With the currvp==0 check, the
     * first two words of data are read on the second line in the frame. The problem
     * occurs when the program jumps to another copperlist a few lines further down
     * which _also_ writes the sprite pointer registers. This means that a) writing
     * to the sprite pointers sets the state to SPR_restart; or b) that sprite DMA
     * is disabled until the end of the vertical blanking interval. The HRM
     * isn't clear - it says that the vertical sprite position can be set to any
     * value, but this wouldn't be the first mistake... */
    /* Update: I modified one of the programs to write the sprite pointers the
     * second time only _after_ the VBlank interval, and it showed the same behaviour
     * as it did unmodified under UAE with the above check. This indicates that the
     * solution below is correct. */
    if (currvp < vblank_endline)
	return;
#endif

    /* The graph in the HRM, p. 195 seems to indicate that sprite 0 is
     * fetched at cycle 0x14 and thus can't be disabled by bitplane DMA. */
    maxspr = currhp/4 - 0x14/4;
    if (maxspr < 0)
	return;
    if (maxspr > NUM_SPRITES)
	maxspr = NUM_SPRITES;

    for (i = 0; i < maxspr; i++) {
	int fetch = 0;

	if (spron[i] == 0)
	    continue;

	if (sprst[i] == SPR_restart) {
	    fetch = 2;
	    spron[i] = 1;
	} else if ((sprst[i] == SPR_waiting_start && sprvstart[i] == vpos) || sprst[i] == SPR_waiting_stop) {
	    fetch = 1;
	    sprst[i] = SPR_waiting_stop;
	}
	if (sprst[i] == SPR_waiting_stop && sprvstop[i] == vpos) {
	    fetch = 2;
	    sprst[i] = SPR_waiting_start;
	}

	if (fetch && dmaen(DMA_SPRITE)) {
	    uae_u16 data1 = chipmem_bank.wget(sprpt[i]);
	    uae_u16 data2 = chipmem_bank.wget(sprpt[i]+real_sprite_width);

	    if (fetch == 1) {
		/* Hack for X mouse auto-calibration */
		if (i == 0 && !sprvbfl && ((sprpos[0] & 0xff) << 2) > 2 * DISPLAY_LEFT_SHIFT) {
		    spr0ctl=sprctl[0];
		    spr0pos=sprpos[0];
		    sprvbfl=2;
		}
		SPRxDATB_1(data2, i);
		SPRxDATA_1(data1, i);
		/* AGA wide sprite hack */
		switch(real_sprite_width)
			{
			case 64>>3:
			data1=chipmem_bank.wget(sprpt[i]+6);
			data2=chipmem_bank.wget(sprpt[i]+6+real_sprite_width);
			SPRxDATB_2(data2, 3*NUM_SPRITES+i);
			SPRxDATA_2(data1, 3*NUM_SPRITES+i);
			data1=chipmem_bank.wget(sprpt[i]+4);
			data2=chipmem_bank.wget(sprpt[i]+4+real_sprite_width);
			SPRxDATB_2(data2, 2*NUM_SPRITES+i);
			SPRxDATA_2(data1, 2*NUM_SPRITES+i);
			case 32>>3:
			data1=chipmem_bank.wget(sprpt[i]+2);
			data2=chipmem_bank.wget(sprpt[i]+2+real_sprite_width);
			SPRxDATB_2(data2, 1*NUM_SPRITES+i);
			SPRxDATA_2(data1, 1*NUM_SPRITES+i);
			break;
		}
	    } else {
		SPRxPOS_1(data1, i);
		SPRxCTL_1(data2, i);
	    }
	    sprpt[i] += real_sprite_width*2;
	}
    }
}

static void adjust_array_sizes (void)
{
#ifdef OS_WITHOUT_MEMORY_MANAGEMENT
    if (delta_sprite_draw) {
	void *p1,*p2;
	int mcc = max_sprite_draw + 200 + delta_sprite_draw;
	delta_sprite_draw = 0;
	p1 = realloc (sprite_positions[0], mcc * sizeof (struct sprite_draw));
	p2 = realloc (sprite_positions[1], mcc * sizeof (struct sprite_draw));
	if (p1) sprite_positions[0] = p1;
	if (p2) sprite_positions[1] = p2;
	if (p1 && p2) {
	    fprintf (stderr, "new max_sprite_draw=%d\n",mcc);
	    max_sprite_draw = mcc;
	}
    }
    if (delta_color_change) {
	void *p1,*p2;
	int mcc = max_color_change + 200 + delta_color_change;
	delta_color_change = 0;
	p1 = realloc (color_changes[0], mcc * sizeof (struct color_change));
	p2 = realloc (color_changes[1], mcc * sizeof (struct color_change));
	if (p1) color_changes[0] = p1;
	if (p2) color_changes[1] = p2;
	if (p1 && p2) {
	    fprintf (stderr, "new max_color_change=%d\n",mcc);
	    max_color_change = mcc;
	}
    }
    if (delta_delay_change) {
	void *p;
	int mcc = max_delay_change + 200 + delta_delay_change;
	delta_delay_change = 0;
	p = realloc (delay_changes, mcc * sizeof (struct delay_change));
	if (p) {
	    fprintf (stderr, "new max_delay_change=%d\n",mcc);
	    delay_changes = p;
	    max_delay_change = mcc;
	}
    }
#endif
}

static void init_hardware_frame (void)
{
    next_lineno = 0;
    nextline_how = nln_normal;
    diwstate = DIW_waiting_start;
    hdiwstate = DIW_waiting_start;
}

void init_hardware_for_drawing_frame (void)
{
    adjust_array_sizes ();

    next_color_change = 0;
    next_delay_change = 0;
    next_sprite_draw = 0;

    next_color_entry = 0;
    remembered_color_entry = -1;
    prev_sprite_positions = sprite_positions[current_change_set];
    curr_sprite_positions = sprite_positions[current_change_set ^ 1];
    prev_color_changes = color_changes[current_change_set];
    curr_color_changes = color_changes[current_change_set ^ 1];
    prev_color_tables = color_tables[current_change_set];
    curr_color_tables = color_tables[current_change_set ^ 1];

    prev_drawinfo = line_drawinfo[current_change_set];
    curr_drawinfo = line_drawinfo[current_change_set ^ 1];
    current_change_set ^= 1;

    color_src_match = color_dest_match = -1;
}

static void vsync_handler (void)
{
    {
        frame_time_t curr_time = read_processor_time ();
        vsyncmintime += vsynctime;
        /* @@@ Mathias? How do you think we should do this? */
        /* If we are too far behind, or we just did a reset, adjust the
         * needed time. */
        if ((long int)(curr_time - vsyncmintime) > 0 || rpt_did_reset)
            vsyncmintime = curr_time + vsynctime;
        rpt_did_reset = 0;
    }

    handle_events ();

    getjoystate (0, &joy0dir, &joy0button);
    getjoystate (1, &joy1dir, &joy1button);

    INTREQ (0x8020);
    if (bplcon0 & 4)
	lof ^= 0x8000;

    vsync_handle_redraw (lof, lof_changed);
    if (quit_program > 0)
	return;

    {
	static int cnt = 0;
	if (cnt == 0) {
	    /* resolution_check_change (); */
	    DISK_check_change ();
	    cnt = 5;
	}
	cnt--;
    }

    lof_changed = 0;

    /* emulate VPOSW refresh rate hacks */
    if(hack_vblank>0&&maxvpos!=hack_vblank) {
	    maxvpos=hack_vblank;
	    vblank_hz=15600/hack_vblank;
	    vsynctime = syncbase / VBLANK_HZ_NTSC;
	    update_sound();
	    write_log("vertical blanking %dHz\n",vblank_hz);
	    hack_vblank=-1;
    } else if(hack_vblank<0) {
	    init_hz();
	    hack_vblank=0;
    }	

    cop_state.ip = cop1lc;
    cop_state.state = COP_read1;
    cop_state.vpos = 0;
    cop_state.hpos = 0;
    cop_state.count = 0;
    cop_state.do_move = 0;
    cop_state.ignore_next = 0;
    cop_state.vdiw = DIW_waiting_start;

    init_hardware_frame ();

#ifdef HAVE_GETTICKCOUNT
    {
	    unsigned long int newtime = GetTickCount() - milliseconds_base;
        unsigned long int delta = newtime - msecs;

	    if (!bogusframe) 
        {
#ifdef SOUND_HAS_PRIORITY
            lastframetime = delta;
            if( n_consecutive_skipped > CUTOFF || lastframetime < 25 ) /* 25ms is the maximum frame-time before we start dropping */
            {
                n_consecutive_skipped = 0;
                clear_inhibit_frame( IHF_SOUNDADJUST );
            }
            else
            {
                n_consecutive_skipped++;
                total_skipped++;
                set_inhibit_frame( IHF_SOUNDADJUST );
            }
#endif
	        frametime += delta;
            cumulative_time += delta;
	        timeframes++;

            if ((timeframes & 127) == 0) /* 128 frames have occured */
            {
		        gui_fps (128000 / cumulative_time ); /* average time of 128 frames */
                cumulative_time = 0;
            }
	    }
	    msecs = newtime;
	    bogusframe = 0;
    }
#endif
    if (ievent_alive > 0)
	ievent_alive--;
    CIA_vsync_handler();

    adjust_sound_timing();
}

static void hsync_handler (void)
{
    update_copper (maxhpos);

    finish_decisions ();
    do_modulos (current_hpos ());

    hsync_record_line_state (next_lineno, nextline_how, thisline_changed);

    eventtab[ev_hsync].evtime += cycles - eventtab[ev_hsync].oldcycles;
    eventtab[ev_hsync].oldcycles = cycles;
    CIA_hsync_handler();

    if (currprefs.produce_sound > 0) {
	int nr;

	update_audio ();

	/* Sound data is fetched at the beginning of each line */
	for (nr = 0; nr < 4; nr++) {
	    struct audio_channel_data *cdp = audio_channel + nr;

	    if (cdp->data_written == 2) {
		cdp->data_written = 0;
		cdp->nextdat = chipmem_bank.wget(cdp->pt);
		cdp->pt += 2;
		if (cdp->state == 2 || cdp->state == 3) {
		    if (cdp->wlen == 1) {
			cdp->pt = cdp->lc;
			cdp->wlen = cdp->len;
			cdp->intreq2 = 1;
		    } else
			cdp->wlen--;
		}
	    }
	}
    }

    hardware_line_completed (next_lineno);

    if (++vpos == (maxvpos + (lof != 0))) {
	vpos = 0;
	vsync_handler();
    }

    is_lastline = ( ( vpos + 1 ) == ( maxvpos + (lof != 0) ) ) && !rpt_did_reset;

    if ((bplcon0 & 4) && currprefs.gfx_linedbl)
	notice_interlace_seen ();

    if (framecnt == 0) {
	    int lineno = vpos;
	    nextline_how = nln_normal;
	    if (currprefs.gfx_linedbl) {
	        lineno *= 2;
	        nextline_how = currprefs.gfx_linedbl == 1 ? nln_doubled : nln_nblack;
	        if (bplcon0 & 4) 
            {
		        if (!lof) 
                {
		            lineno++;
		            nextline_how = nln_lower;
		        }
                else
                {
    		        nextline_how = nln_upper;
	    	    }
	        }
	    }
        next_lineno = lineno;
	    reset_decisions ();
    }
    if (uae_int_requested) {
	set_uae_int_flag ();
	INTREQ (0xA000);
    }
    update_copper (0);
    prepare_copper ();
    if (eventtab[ev_copper].evtime == cycles && eventtab[ev_copper].active)
	abort ();
}

static void init_eventtab (void)
{
    int i;

    for(i = 0; i < ev_max; i++) {
	eventtab[i].active = 0;
	eventtab[i].oldcycles = 0;
    }

    eventtab[ev_cia].handler = CIA_handler;
    eventtab[ev_copper].handler = do_copper;
    eventtab[ev_hsync].handler = hsync_handler;
    eventtab[ev_hsync].evtime = maxhpos + cycles;
    eventtab[ev_hsync].active = 1;

    eventtab[ev_blitter].handler = blitter_handler;
    eventtab[ev_blitter].active = 0;
    eventtab[ev_diskblk].handler = diskblk_handler;
    eventtab[ev_diskblk].active = 0;
    eventtab[ev_diskindex].handler = diskindex_handler;
    eventtab[ev_diskindex].active = 0;

    events_schedule ();
}

void customreset (void)
{
    int i;
    uae_u16 v;
    uae_u32 vv;

    for (i = 0; i < sizeof current_colors.color_regs / sizeof *current_colors.color_regs; i++) 
    {
	    if(!savestate_wanted)
            current_colors.color_regs[i] = -1;
    }

    expamem_reset ();

    CIA_reset ();
    cycles = 0;
    regs.spcflags &= SPCFLAG_BRK;

    vpos = 0;
    lof = 0;

    if (needmousehack ()) 
    {
    #if 0
	    mousehack_setfollow();
    #else
	    mousehack_setdontcare();
    #endif
    }
    else
    {
	    mousestate = normal_mouse;
    }
    ievent_alive = 0;

    cycles = 0;
    bltstate = BLT_done;
    diwstate = DIW_waiting_start;
    hdiwstate = DIW_waiting_start;

    if (!savestate_wanted) 
    {
	    clx_sprmask = 0xFF;
	    clxdat = 0;

	    memset (sprarmed, 0, sizeof sprarmed);
	    nr_armed = 0;

	    dmacon = intena = 0;
	    cop_state.state = COP_stop;
	    copcon = 0;
	    dskdmaen = 0;

	    bplcon4 = 0x11; /* Get AGA currprefs.chipset into ECS compatibility mode */
	    bplcon3 = 0xC00;
	    beamcon0 = currprefs.ntscmode ? 0x00 : 0x20;
	    audio_reset ();
	    for (i = 0; i < NUM_SPRITES; i++) 
        {
	        /* ???? */
	        sprst[i] = SPR_stop;
	        spron[i] = 0;
	    }
	    memset (sprpos, 0, sizeof sprpos);
	    memset (sprctl, 0, sizeof sprctl);
	    real_sprite_width=16>>3;
	    used_sprites=8;
    }
    init_hz ();

    init_eventtab ();

    init_hardware_frame ();
    reset_drawing ();

    reset_decisions ();

#ifdef HAVE_GETTICKCOUNT
    milliseconds_base = GetTickCount();
    bogusframe = 1;
#endif
    if (savestate_wanted) 
    {
	    ADKCON(0);
	    INTENA(0);
	    INTREQ(0);
	    DMACON(0);
	    COPJMP1(0);
	    if(diwhigh) diwhigh_written=1;
	    calcdiw();
	    v=BPLCON0_R();
	    BPLCON0_W(0);
	    BPLCON0(0,v);
	    for(i=0;i<256;i++) 
        {
		    vv=current_colors.color_regs[i];
		    current_colors.color_regs[i]=-1;
		    record_color_change (0, i, vv);
		    remembered_color_entry = -1;
		    current_colors.color_regs[i] = vv;
		    if(i<32&&!(currprefs.chipset&AGA_CHIPSET)) current_colors.acolors[i] = xcolors[vv];
	    }
	    write_log("State restored\n");
	    savestate_wanted=0;
    }
}

void dumpcustom (void)
{
    write_log ("DMACON: %x INTENA: %x INTREQ: %x VPOS: %x HPOS: %x\n", DMACONR(),
	     (unsigned int)intena, (unsigned int)intreq, (unsigned int)vpos, (unsigned int)current_hpos());
    write_log ("COP1LC: %08lx, COP2LC: %08lx\n", (unsigned long)cop1lc, (unsigned long)cop2lc);
    if (timeframes) {
	write_log ("Average frame time: %d ms [frames: %d time: %d]\n",
		 frametime/timeframes, timeframes, frametime);
    }
#ifdef SOUND_HAS_PRIORITY
    if (total_skipped) {
        write_log ("Dropped frames: %u\n", total_skipped);
    }
#endif
    dump_audio_bench ();
    /*for (i=0; i<256; i++) if (blitcount[i]) fprintf (stderr, "minterm %x = %d\n",i,blitcount[i]);  blitter debug */
}

int intlev (void)
{
    uae_u16 imask = intreq & intena;
    if (imask && (intena & 0x4000)){
	if (imask & 0x2000) return 6;
	if (imask & 0x1800) return 5;
	if (imask & 0x0780) return 4;
	if (imask & 0x0070) return 3;
	if (imask & 0x0008) return 2;
	if (imask & 0x0007) return 1;
    }
    return -1;
}

static void gen_custom_tables (void)
{
    int i;
    for (i = 0; i < 256; i++) {
	unsigned int j;
	sprtaba[i] = ((((i >> 7) & 1) << 0)
		      | (((i >> 6) & 1) << 2)
		      | (((i >> 5) & 1) << 4)
		      | (((i >> 4) & 1) << 6)
		      | (((i >> 3) & 1) << 8)
		      | (((i >> 2) & 1) << 10)
		      | (((i >> 1) & 1) << 12)
		      | (((i >> 0) & 1) << 14));
	sprtabb[i] = sprtaba[i] * 2;
	for (j = 0; j < 511; j = (j << 1) | 1)
	    if ((i & ~j) == 0)
		waitmasktab[i] = ~j;
    }
}

void custom_init (void)
{
#ifdef OS_WITHOUT_MEMORY_MANAGEMENT
    int num;

    for (num = 0; num < 2; num++) {
       sprite_positions[num] = xmalloc (max_sprite_draw * sizeof (struct sprite_draw));
       color_changes[num] = xmalloc (max_color_change * sizeof (struct color_change));
    }
    delay_changes = xmalloc (max_delay_change * sizeof (struct delay_change));
#endif
    init_decisions ();

    gen_custom_tables ();
    build_blitfilltable ();

    drawing_init ();

    mousestate = unknown_mouse;

    if (needmousehack ())
	mousehack_setfollow();
}

void custom_cleanup( void )
{
    drawing_cleanup();
}

/* Custom chip memory bank */

static uae_u32 custom_lget (uaecptr) REGPARAM;
static uae_u32 custom_wget (uaecptr) REGPARAM;
static uae_u32 custom_bget (uaecptr) REGPARAM;
static void custom_lput (uaecptr, uae_u32) REGPARAM;
static void custom_wput (uaecptr, uae_u32) REGPARAM;
static void custom_bput (uaecptr, uae_u32) REGPARAM;

addrbank custom_bank = {
    custom_lget, custom_wget, custom_bget,
    custom_lput, custom_wput, custom_bput,
    default_xlate, default_check
};

uae_u32 REGPARAM2 custom_wget (uaecptr addr)
{
    switch (addr & 0x1FE) {
     case 0x002: return DMACONR();
     case 0x004: return VPOSR();
     case 0x006: return VHPOSR();

     case 0x008: return DSKDATR();

     case 0x00A: return JOY0DAT();
     case 0x00C: return JOY1DAT();
     case 0x00E: return CLXDAT();
     case 0x010: return ADKCONR();

     case 0x012: return POT0DAT();
     case 0x016: return POTGOR();
     case 0x018: return SERDATR();
     case 0x01A: return DSKBYTR();
     case 0x01C: return INTENAR();
     case 0x01E: return INTREQR();
     case 0x07c: return LISAID();
 
     case 0x180: case 0x182: case 0x184: case 0x186: case 0x188: case 0x18A:
     case 0x18C: case 0x18E: case 0x190: case 0x192: case 0x194: case 0x196:
     case 0x198: case 0x19A: case 0x19C: case 0x19E: case 0x1A0: case 0x1A2:
     case 0x1A4: case 0x1A6: case 0x1A8: case 0x1AA: case 0x1AC: case 0x1AE:
     case 0x1B0: case 0x1B2: case 0x1B4: case 0x1B6: case 0x1B8: case 0x1BA:
     case 0x1BC: case 0x1BE:
	return COLOR_READ ((addr & 0x3E) / 2);
	break;

    default:
         custom_wput(addr,0);
	return 0xffff;
    }
}

uae_u32 REGPARAM2 custom_bget (uaecptr addr)
{
    return custom_wget(addr & 0xfffe) >> (addr & 1 ? 0 : 8);
}

uae_u32 REGPARAM2 custom_lget (uaecptr addr)
{
    return ((uae_u32)custom_wget(addr & 0xfffe) << 16) | custom_wget((addr+2) & 0xfffe);
}

void REGPARAM2 custom_wput_1 (int hpos, uaecptr addr, uae_u32 value)
{
    addr &= 0x1FE;
    switch (addr) {
     case 0x020: DSKPTH (value); break;
     case 0x022: DSKPTL (value); break;
     case 0x024: DSKLEN (value); break;
     case 0x026: DSKDAT (value); break;

     case 0x02A: VPOSW (value); break;
     case 0x2E:  COPCON (value); break;
     case 0x030: SERDAT (value); break;
     case 0x032: SERPER (value); break;
     case 0x34: POTGO (value); break;
     case 0x040: BLTCON0 (value); break;
     case 0x042: BLTCON1 (value); break;

     case 0x044: BLTAFWM (value); break;
     case 0x046: BLTALWM (value); break;

     case 0x050: BLTAPTH (value); break;
     case 0x052: BLTAPTL (value); break;
     case 0x04C: BLTBPTH (value); break;
     case 0x04E: BLTBPTL (value); break;
     case 0x048: BLTCPTH (value); break;
     case 0x04A: BLTCPTL (value); break;
     case 0x054: BLTDPTH (value); break;
     case 0x056: BLTDPTL (value); break;

     case 0x058: BLTSIZE (value); break;

     case 0x064: BLTAMOD (value); break;
     case 0x062: BLTBMOD (value); break;
     case 0x060: BLTCMOD (value); break;
     case 0x066: BLTDMOD (value); break;

     case 0x070: BLTCDAT (value); break;
     case 0x072: BLTBDAT (value); break;
     case 0x074: BLTADAT (value); break;

     case 0x07E: DSKSYNC (value); break;

     case 0x080: COP1LCH (value); break;
     case 0x082: COP1LCL (value); break;
     case 0x084: COP2LCH (value); break;
     case 0x086: COP2LCL (value); break;

     case 0x088: COPJMP1 (value); break;
     case 0x08A: COPJMP2 (value); break;

     case 0x08E: DIWSTRT (hpos, value); break;
     case 0x090: DIWSTOP (hpos, value); break;
     case 0x092: DDFSTRT (hpos, value); break;
     case 0x094: DDFSTOP (hpos, value); break;

     case 0x096: DMACON (value); break;
     case 0x098: CLXCON (value); break;
     case 0x09A: INTENA (value); break;
     case 0x09C: INTREQ (value); break;
     case 0x09E: ADKCON (value); break;

     case 0x0A0: AUDxLCH (0, value); break;
     case 0x0A2: AUDxLCL (0, value); break;
     case 0x0A4: AUDxLEN (0, value); break;
     case 0x0A6: AUDxPER (0, value); break;
     case 0x0A8: AUDxVOL (0, value); break;
     case 0x0AA: AUDxDAT (0, value); break;

     case 0x0B0: AUDxLCH (1, value); break;
     case 0x0B2: AUDxLCL (1, value); break;
     case 0x0B4: AUDxLEN (1, value); break;
     case 0x0B6: AUDxPER (1, value); break;
     case 0x0B8: AUDxVOL (1, value); break;
     case 0x0BA: AUDxDAT (1, value); break;

     case 0x0C0: AUDxLCH (2, value); break;
     case 0x0C2: AUDxLCL (2, value); break;
     case 0x0C4: AUDxLEN (2, value); break;
     case 0x0C6: AUDxPER (2, value); break;
     case 0x0C8: AUDxVOL (2, value); break;
     case 0x0CA: AUDxDAT (2, value); break;

     case 0x0D0: AUDxLCH (3, value); break;
     case 0x0D2: AUDxLCL (3, value); break;
     case 0x0D4: AUDxLEN (3, value); break;
     case 0x0D6: AUDxPER (3, value); break;
     case 0x0D8: AUDxVOL (3, value); break;
     case 0x0DA: AUDxDAT (3, value); break;

     case 0x0E0: BPLPTH (hpos, value, 0); break;
     case 0x0E2: BPLPTL (hpos, value, 0); break;
     case 0x0E4: BPLPTH (hpos, value, 1); break;
     case 0x0E6: BPLPTL (hpos, value, 1); break;
     case 0x0E8: BPLPTH (hpos, value, 2); break;
     case 0x0EA: BPLPTL (hpos, value, 2); break;
     case 0x0EC: BPLPTH (hpos, value, 3); break;
     case 0x0EE: BPLPTL (hpos, value, 3); break;
     case 0x0F0: BPLPTH (hpos, value, 4); break;
     case 0x0F2: BPLPTL (hpos, value, 4); break;
     case 0x0F4: BPLPTH (hpos, value, 5); break;
     case 0x0F6: BPLPTL (hpos, value, 5); break;
     case 0x0F8: BPLPTH (hpos, value, 6); break;
     case 0x0FA: BPLPTL (hpos, value, 6); break;
     case 0x0FC: BPLPTH (hpos, value, 7); break;
     case 0x0FE: BPLPTL (hpos, value, 7); break;

     case 0x100: BPLCON0 (hpos, value); break;
     case 0x102: BPLCON1 (hpos, value); break;
     case 0x104: BPLCON2 (hpos, value); break;
     case 0x106: BPLCON3 (hpos, (currprefs.chipset & AGA_CHIPSET) ? value : value & 0xff); break;

     case 0x108: BPL1MOD (hpos, value); break;
     case 0x10A: BPL2MOD (hpos, value); break;

     case 0x110: BPL1DAT (value); break;
     case 0x112: BPL2DAT (value); break;
     case 0x114: BPL3DAT (value); break;
     case 0x116: BPL4DAT (value); break;
     case 0x118: BPL5DAT (value); break;
     case 0x11A: BPL6DAT (value); break;
     case 0x11C: BPL7DAT (value); break;
     case 0x11E: BPL8DAT (value); break;

     case 0x180: case 0x182: case 0x184: case 0x186: case 0x188: case 0x18A:
     case 0x18C: case 0x18E: case 0x190: case 0x192: case 0x194: case 0x196:
     case 0x198: case 0x19A: case 0x19C: case 0x19E: case 0x1A0: case 0x1A2:
     case 0x1A4: case 0x1A6: case 0x1A8: case 0x1AA: case 0x1AC: case 0x1AE:
     case 0x1B0: case 0x1B2: case 0x1B4: case 0x1B6: case 0x1B8: case 0x1BA:
     case 0x1BC: case 0x1BE:
	COLOR_WRITE (hpos, value & 0xFFF, (addr & 0x3E) / 2);
	break;
     case 0x120: case 0x124: case 0x128: case 0x12C:
     case 0x130: case 0x134: case 0x138: case 0x13C:
	SPRxPTH (hpos, value, (addr - 0x120) / 4);
	break;
     case 0x122: case 0x126: case 0x12A: case 0x12E:
     case 0x132: case 0x136: case 0x13A: case 0x13E:
	SPRxPTL (hpos, value, (addr - 0x122) / 4);
	break;
     case 0x140: case 0x148: case 0x150: case 0x158:
     case 0x160: case 0x168: case 0x170: case 0x178:
	SPRxPOS (hpos, value, (addr - 0x140) / 8);
	break;
     case 0x142: case 0x14A: case 0x152: case 0x15A:
     case 0x162: case 0x16A: case 0x172: case 0x17A:
	SPRxCTL (hpos, value, (addr - 0x142) / 8);
	break;
     case 0x144: case 0x14C: case 0x154: case 0x15C:
     case 0x164: case 0x16C: case 0x174: case 0x17C:
	SPRxDATA (hpos, value, (addr - 0x144) / 8);
	break;
     case 0x146: case 0x14E: case 0x156: case 0x15E:
     case 0x166: case 0x16E: case 0x176: case 0x17E:
	SPRxDATB (hpos, value, (addr - 0x146) / 8);
	break;

     case 0x36: JOYTEST (value); break;
     case 0x5A: if(currprefs.chipset & (AGA_CHIPSET|ECS_AGNUS)) BLTCON0L(value); break;
     case 0x5C: if(currprefs.chipset & (AGA_CHIPSET|ECS_AGNUS)) BLTSIZV (value); break;
     case 0x5E: if(currprefs.chipset & (AGA_CHIPSET|ECS_AGNUS)) BLTSIZH (value); break;

     case 0x1E4: if(currprefs.chipset & (AGA_CHIPSET|ECS_DENISE)) DIWHIGH (hpos, value); break;

     case 0x10C: if(currprefs.chipset & AGA_CHIPSET) BPLCON4 (hpos, value); break;
     case 0x1DC: if(currprefs.chipset & (AGA_CHIPSET|ECS_AGNUS)) BEAMCON0 (value); break;
     case 0x1FC: if(currprefs.chipset & AGA_CHIPSET) FMODE(value); break;
    }
}

void REGPARAM2 custom_wput (uaecptr addr, uae_u32 value)
{
    int hpos = current_hpos ();
    update_copper (hpos);
    custom_wput_1 (hpos, addr, value);
}

void REGPARAM2 custom_bput (uaecptr addr, uae_u32 value)
{
    static int warned = 0;
    /* Is this correct now? (There are people who bput things to the upper byte of AUDxVOL). */
    uae_u16 rval = (value << 8) | (value & 0xFF);
    custom_wput(addr, rval);
    if (!warned)
	write_log ("Byte put to custom register.\n"), warned++;
}

void REGPARAM2 custom_lput(uaecptr addr, uae_u32 value)
{
    custom_wput(addr & 0xfffe, value >> 16);
    custom_wput((addr+2) & 0xfffe, (uae_u16)value);
}

/* Custom currprefs.chipset save/restore code */

void custom_save_state(void)
{
char str[1000];
int len, olddmacon;

if(dskdmaen>1) {
	write_log("Waiting for disk DMA to stop...\n"); /* No support for saving Disk DMA state yet */
	return;
}
/* force blitter to finish, no support for saving blitter state yet */
if (eventtab[ev_blitter].active) {
	olddmacon=dmacon;
	dmacon|=DMA_BLITTER; /* ugh.. */
	blitter_handler();
	dmacon=olddmacon;
}
 /* FIXME: add dialog for asking state file name */
strcpy(str,currprefs.df[0]);
len=strlen(str);
len-=4;
if(len>=0) strcpy(str+len,".asf");
/* and description... */
save_state(str,"hihhui... test savestate file..");
savestate_wanted=0;
}

uae_u8 *restore_custom(uae_u8 *src)
{
int i;

audio_reset();

currprefs.chipset=restore_long();
restore_word();			// 000 ?
restore_word();			// 002 DMACONR
restore_word();			// 004 VPOSR
restore_word();			// 006 VHPOSR
restore_word();			// 008 DSKDATR
restore_word();			// 00A JOY0DAT
restore_word();			// 00C JOY1DAT
CLXDAT_W(restore_word());	// 00E CLXDAT
restore_word();			// 010 ADKCONR
restore_word();			// 012 POT0DAT *
restore_word();			// 014 POT1DAT *
restore_word();			// 016 POTINP *
restore_word();			// 018 SERDATR *
restore_word();			// 01A DSKBYTR *
restore_word();			// 01C INTENAR
restore_word();			// 01E INTREQR
DSKPTH_W(restore_word());	// 020 DSKPTH
DSKPTL_W(restore_word());	// 022 DSKPTL
DSKLEN_W(restore_word());	// 024 DSKLEN
restore_word();			// 026 DSKDAT
restore_word();			// 028 REFPTR
VPOSW_W(restore_word());	// 02A VPOSW
restore_word();			// 02C VHPOSW
COPCON_W(restore_word());	// 02E COPCON
restore_word();			// 030 SERDAT *
restore_word();			// 032 SERPER *
POTGO_W(restore_word());// 034 POTGO *
restore_word();			// 036 JOYTEST *
restore_word();			// 038 STREQU
restore_word();			// 03A STRVHBL
restore_word();			// 03C STRHOR
restore_word();			// 03E STRLONG
BLTCON0_W(restore_word());	// 040 BLTCON0
BLTCON1_W(restore_word());	// 042 BLTCON1
BLTAFWM_W(restore_word());	// 044 BLTAFWM
BLTALWM_W(restore_word());	// 046 BLTALWM
BLTCPTH_W(restore_word());	// 048 BLTCPTH
BLTCPTL_W(restore_word());	// 04A BLTCPTL
BLTBPTH_W(restore_word());	// 04C BLTBPTH
BLTBPTL_W(restore_word());	// 04E BLTBPTL
BLTAPTH_W(restore_word());	// 050 BLTAPTH
BLTAPTL_W(restore_word());	// 052 BLTAPTL
BLTDPTH_W(restore_word());	// 054 BLTDPTH
BLTDPTL_W(restore_word());	// 056 BLTDPTL
BLTSIZE_W(restore_word());	// 058 BLTSIZE
BLTCON0L_W(restore_word());	// 05A BLTCON0L
BLTSIZV_W(restore_word());	// 05C BLTSIZV
BLTSIZH_W(restore_word());	// 05E BLTSIZH
BLTCMOD_W(restore_word());	// 060 BLTCMOD
BLTBMOD_W(restore_word());	// 062 BLTBMOD
BLTAMOD_W(restore_word());	// 064 BLTAMOD
BLTDMOD_W(restore_word());	// 066 BLTDMOD
restore_word();			// 068 ?
restore_word();			// 06A ?
restore_word();			// 06C ?
restore_word();			// 06E ?
BLTCDAT_W(restore_word());	// 070 BLTCDAT
BLTBDAT_W(restore_word());	// 072 BLTBDAT
BLTADAT_W(restore_word());	// 074 BLTADAT
restore_word();			// 076 ?
restore_word();			// 078 ?
restore_word();			// 07A ?
restore_word();			// 07C LISAID
DSKSYNC_W(restore_word());	// 07E DSKSYNC
COP1LCH_W(restore_word());	// 080 COP1LCH
COP1LCL_W(restore_word());	// 082 COP1LCL
COP2LCH_W(restore_word());	// 084 COP2LCH
COP2LCL_W(restore_word());	// 086 COP2LCL
restore_word();			// 088 ?
restore_word();			// 08A ?
restore_word();			// 08C ?
DIWSTRT_W(restore_word());	// 08E DIWSTRT
DIWSTOP_W(restore_word());	// 090 DIWSTOP
DDFSTRT_W(restore_word());	// 092 DDFSTRT
DDFSTOP_W(restore_word());	// 094 DDFSTOP
DMACON_W(restore_word());	// 096 DMACON
CLXCON_W(restore_word());	// 098 CLXCON
INTENA_W(restore_word());	// 09A INTENA
INTREQ_W(restore_word());	// 09C INTREQ
ADKCON_W(restore_word());	// 09E ADKCON
src=restore_audio(src,0);
for(i=0;i<8;i++) {
    BPLPTH_W(i,restore_word());	// 0E0 BPLxPTH
    BPLPTL_W(i,restore_word());	// 0E2 BPLxPTL
}
BPLCON0_W(restore_word());	// 100 BPLCON0
BPLCON1_W(restore_word());	// 102 BPLCON1
BPLCON2_W(restore_word());	// 104 BPLCON2
BPLCON3_W(restore_word());	// 106 BPLCON3
BPL1MOD_W(restore_word());	// 108 BPL1MOD
BPL2MOD_W(restore_word());	// 10A BPL2MOD
BPLCON4_W(restore_word());	// 10C BPLCON4
CLXCON2_W(restore_word());	// 10E CLXCON2
for(i=0;i<8;i++) {
     restore_word();		// 110 BPLxDAT
}
for(i=0;i<8;i++) {
     SPRxPTH_W(i,restore_word());// 120 SPRxPTH
     SPRxPTL_W(i,restore_word());// 122 SPRxPTL
}
for(i=0;i<8;i++) {
     SPRxPOS_W(i,restore_word());// 140 SPRxPOS
     SPRxCTL_W(i,restore_word());// 142 SPRxCTL
     SPRxDATA_W(i,restore_word());// 144 SPRxDATA
     SPRxDATB_W(i,restore_word());// 146 SPRxDATB
}
for(i=0;i<32;i++) {
     COLOR_W(i,restore_word());	// 180 COLORxx
}
restore_word();			// 1C0
restore_word();			// 1C2
restore_word();			// 1C4
restore_word();			// 1C6
restore_word();			// 1C8
restore_word();			// 1CA
restore_word();			// 1CC
restore_word();			// 1CE
restore_word();			// 1D0
restore_word();			// 1D2
restore_word();			// 1D4
restore_word();			// 1D6
restore_word();			// 1D8
restore_word();			// 1DA
BEAMCON0_W(restore_word());	// 1DC
restore_word();			// 1DE
restore_word();			// 1E0
restore_word();			// 1E2
restore_word();			// 1E4
restore_word();			// 1E6
restore_word();			// 1E8
restore_word();			// 1EA
restore_word();			// 1EC
restore_word();			// 1EE
restore_word();			// 1F0
restore_word();			// 1F2
restore_word();			// 1F4
restore_word();			// 1F6
restore_word();			// 1F8
restore_word();			// 1FA
real_sprite_width=16>>3;
used_sprites=8;
FMODE_W(restore_word());	// 1FC
restore_word();			// 1FE

for(i=0;i<256;i++) COLOR_W(i,restore_long());

for(i=0;i<8;i++) sprarmed[i]=restore_byte()&1;

src=restore_audio(src,1);

return src;
}

uae_u8 *save_custom(int *len)
{
uae_u8 *dstbak, *dst;
int i;

dstbak=dst=xmalloc(8+512+256*4+200);
save_long(currprefs.chipset);
save_word(0);			// 000 ?
save_word(DMACONR());		// 002 DMACONR
save_word(VPOSR());		// 004 VPOSR
save_word(VHPOSR());		// 006 VHPOSR
save_word(DSKDATR());		// 008 DSKDATR
save_word(JOY0DAT());		// 00A JOY0DAT
save_word(JOY1DAT());		// 00C JOY1DAT
save_word(CLXDAT_R());		// 00E CLXDAT
save_word(ADKCONR());		// 010 ADKCONR
save_word(POT0DAT());		// 012 POT0DAT
save_word(POT0DAT());		// 014 POT1DAT
save_word(0)	;		// 016 POTINP *
save_word(0);			// 018 SERDATR *
save_word(DSKBYTR());		// 01A DSKBYTR
save_word(INTENAR());		// 01C INTENAR
save_word(INTREQR());		// 01E INTREQR
save_word(DSKPTH_R());		// 020 DSKPTH
save_word(DSKPTL_R());		// 022 DSKPTL
save_word(DSKLEN_R());		// 024 DSKLEN
save_word(DSKDAT_R());		// 026 DSKDAT
save_word(0);			// 028 REFPTR
save_word(VPOSW_R());		// 02A VPOSW
save_word(0);			// 02C VHPOSW
save_word(COPCON_R());		// 02E COPCON
save_word(0);			// 030 SERDAT *
save_word(0);			// 032 SERPER *
save_word(POTGO_R());		// 034 POTGO
save_word(JOYTEST_R());		// 036 JOYTEST *
save_word(0);			// 038 STREQU
save_word(0);			// 03A STRVBL
save_word(0);			// 03C STRHOR
save_word(0);			// 03E STRLONG
save_word(BLTCON0_R());		// 040 BLTCON0
save_word(BLTCON1_R());		// 042 BLTCON1
save_word(BLTAFWM_R());		// 044 BLTAFWM
save_word(BLTALWM_R());		// 046 BLTALWM
save_word(BLTCPTH_R());		// 048 BLTCPTH
save_word(BLTCPTL_R());		// 04A BLTCPTL
save_word(BLTBPTH_R());		// 04C BLTBPTH
save_word(BLTBPTL_R());		// 04E BLTBPTL
save_word(BLTAPTH_R());		// 050 BLTAPTH
save_word(BLTAPTL_R());		// 052 BLTAPTL
save_word(BLTDPTH_R());		// 054 BLTDPTH
save_word(BLTDPTL_R());		// 056 BLTDPTL
save_word(BLTSIZE_R());		// 058 BLTSIZE
save_word(BLTCON0L_R());	// 05A BLTCON0L
save_word(BLTSIZV_R());		// 05C BLTSIZV
save_word(BLTSIZH_R());		// 05E BLTSIZH
save_word(BLTCMOD_R());		// 060 BLTCMOD
save_word(BLTBMOD_R());		// 062 BLTBMOD
save_word(BLTAMOD_R());		// 064 BLTAMOD
save_word(BLTDMOD_R());		// 066 BLTDMOD
save_word(0);			// 068 ?
save_word(0);			// 06A ?
save_word(0);			// 06C ?
save_word(0);			// 06E ?
save_word(BLTCDAT_R());		// 070 BLTCDAT
save_word(BLTBDAT_R());		// 072 BLTBDAT
save_word(BLTADAT_R());		// 074 BLTADAT
save_word(0);			// 076 ?
save_word(0);			// 078 ?
save_word(0);			// 07A ?
save_word(LISAID());		// 07C LISAID
save_word(DSKSYNC_R());		// 07E DSKSYNC
save_word(COP1LCH_R());		// 080 COP1LCH
save_word(COP1LCL_R());		// 082 COP1LCL
save_word(COP2LCH_R());		// 084 COP2LCH
save_word(COP2LCL_R());		// 086 COP2LCL
save_word(0);			// 088 ?
save_word(0);			// 08A ?
save_word(0);			// 08C ?
save_word(DIWSTRT_R());		// 08E DIWSTRT
save_word(DIWSTOP_R());		// 090 DIWSTOP
save_word(DDFSTRT_R());		// 092 DDFSTRT
save_word(DDFSTOP_R());		// 094 DDFSTOP
save_word(DMACON_R());		// 096 DMACON
save_word(CLXCON_R());		// 098 CLXCON
save_word(INTENA_R());		// 09A INTENA
save_word(INTREQ_R());		// 09C INTREQ
save_word(ADKCON_R());		// 09E ADKCON
dst=save_audio(dst,0);
for(i=0;i<8;i++) {
    save_word(BPLPTH_R(i));	// 0E0 BPLxPTH
    save_word(BPLPTL_R(i));	// 0E2 BPLxPTL
}
save_word(BPLCON0_R());		// 100 BPLCON0
save_word(BPLCON1_R());		// 102 BPLCON1
save_word(BPLCON2_R());		// 104 BPLCON2
save_word(BPLCON3_R());		// 106 BPLCON3
save_word(BPL1MOD_R());		// 108 BPL1MOD
save_word(BPL2MOD_R());		// 10A BPL2MOD
save_word(BPLCON4_R());		// 10C BPLCON4
save_word(CLXCON2_R());		// 10E CLXCON2
for(i=0;i<8;i++) {
     save_word(0);		// 110 BPLxDAT
}
for(i=0;i<8;i++) {
     save_word(SPRxPTH_R(i));	// 120 SPRxPTH
     save_word(SPRxPTL_R(i));	// 122 SPRxPTL
}
for(i=0;i<8;i++) {
     save_word(SPRxPOS_R(i));	// 140 SPRxPOS
     save_word(SPRxCTL_R(i));	// 142 SPRxCTL
     save_word(SPRxDATA_R(i));	// 144 SPRxDATA
     save_word(SPRxDATB_R(i));	// 146 SPRxDATB
}
for(i=0;i<32;i++) {
     save_word((uae_u16)COLOR_R(i));// 180 COLORxx
}
save_word(0);			// 1C0
save_word(0);			// 1C2
save_word(0);			// 1C4
save_word(0);			// 1C6
save_word(0);			// 1C8
save_word(0);			// 1CA
save_word(0);			// 1CC
save_word(0);			// 1CE
save_word(0);			// 1D0
save_word(0);			// 1D2
save_word(0);			// 1D4
save_word(0);			// 1D6
save_word(0);			// 1D8
save_word(0);			// 1DA
save_word(BEAMCON0_R());	// 1DC
save_word(0);			// 1DE
save_word(0);			// 1E0
save_word(0);			// 1E2
save_word(0);			// 1E4
save_word(0);			// 1E6
save_word(0);			// 1E8
save_word(0);			// 1EA
save_word(0);			// 1EC
save_word(0);			// 1EE
save_word(0);			// 1F0
save_word(0);			// 1F2
save_word(0);			// 1F4
save_word(0);			// 1F6
save_word(0);			// 1F8
save_word(0);			// 1FA
save_word(FMODE_R());		// 1FC
save_word(0xffff);		// 1FE

/* AGA colors */
for(i=0;i<256;i++) save_long(COLOR_R(i));

/* sprites' armed status (for non-DMA mode sprites) */
for(i=0;i<8;i++) save_byte(sprarmed[i]?1:0);

/* internal audio state */
dst=save_audio(dst,1);

*len=dst-dstbak;
return dstbak;
}
