 /*
  * UAE - The Un*x Amiga Emulator
  *
  * Save/restore emulator state
  *
  * (c) 1999 Toni Wilen
  *
  * see below for ASF-structure
  */

 /* Features:
  *
  * - full CPU state (68000/68010/68020, no FPU)
  * - full CIA-A and CIA-B state (with all internal registers)
  * - all custom registers and audio internal state. NOTE: not all registers are restored yet.
  * - only chip memory and bogo memory is saved and restored. Fast memory will be supported later.
  * - disk drive type, imagefile, track and motor state
  * - Kickstart ROM version, address and size is saved. This data is not used during restore yet.
  */

 /* Notes:
  *
  * - blitter state is not saved, blitter is forced to finish immediately if it
  *   was active
  * - disk DMA state is not completely saved. Currently we just delay saving
  *   until DMA has completed.
  * - does not ask for statefile name and description. Currently uses DF0's disk
  *   image name (".adf" is replaced with ".asf") WinUAE: Brian, add dialog box :)
  * - only Amiga state is restored, harddisk support, expansion boards etc..
  *   are not saved/restored.
  * - se this for saving games that do not have real save support or use strange
  *   save format.
  */

 /* Usage :
  *
  * save:
  * 
  * set "savestate_wanted" = 1. During every vertical blank "savestate_wanted"
  * is checked and if it is != 0, custom_save_state() is called which in turn
  * calls save_state() if disk DMA was not active.
  *
  * restore:
  * 
  * just call restore_state(<filename>) after keybuf_init() in main.c.
  *
  */

#include "sysconfig.h"
#include "sysdeps.h"

#include "config.h"
#include "options.h"

#include "savestate.h"

int savestate_wanted;

/* functions for reading/writing bytes, shorts and longs in big-endian
 * format independent of host machine's endianess */

void save_long_func(uae_u8 **dstp, uae_u32 v)
{
uae_u8 *dst=*dstp;
*dst++=(uae_u8)(v>>24);
*dst++=(uae_u8)(v>>16);
*dst++=(uae_u8)(v>>8);
*dst++=(uae_u8)(v>>0);
*dstp=dst;
}
void save_word_func(uae_u8 **dstp, uae_u16 v)
{
uae_u8 *dst=*dstp;
*dst++=(uae_u8)(v>>8);
*dst++=(uae_u8)(v>>0);
*dstp=dst;
}
void save_byte_func(uae_u8 **dstp, uae_u8 v)
{
uae_u8 *dst=*dstp;
*dst++=v;
*dstp=dst;
}

uae_u32 restore_long_func(uae_u8 **dstp)
{
uae_u32 v;
uae_u8 *dst=*dstp;
v=(dst[0]<<24)|(dst[1]<<16)|(dst[2]<<8)|(dst[3]);
*dstp=dst+4;
return v;
}
uae_u16 restore_word_func(uae_u8 **dstp)
{
uae_u16 v;
uae_u8 *dst=*dstp;
v=(dst[0]<<8)|(dst[1]);
*dstp=dst+2;
return v;
}
uae_u8 restore_byte_func(uae_u8 **dstp)
{
uae_u8 v;
uae_u8 *dst=*dstp;
v=dst[0];
*dstp=dst+1;
return v;
}

/* read and write IFF-style hunks */

static void save_chunk(FILE *f,uae_u8 *chunk,int len,char *name)
{
uae_u8 size[4],*dst;
uae_u8 zero[4]={0,0,0,0};

fwrite(name,1,4,f);
dst=&size[0];
save_long(len+4+4);
fwrite(&size[0],1,4,f);
fwrite(chunk,1,len,f);
len=4-(len&3);
if(len) fwrite(zero,1,len,f);
}

static uae_u8 *restore_chunk(FILE *f,char *name,int *len)
{
uae_u8 size[4],dummy[4],*mem,*src;
int len2;

fread(name,1,4,f);
name[4]=0;
fread(size,1,4,f);
src=size;
len2=restore_long()-4-4;
if(len2<=0) return 0;
mem=malloc(len2);
fread(mem,1,len2,f);
*len=len2;
len2=4-(len2&3);
if(len2) fread(dummy,1,len2,f);
return mem;
}

static void restore_header(uae_u8 *src)
{
char *emuname,*emuversion,*description;

restore_long();
emuname=src;
src+=strlen(src)+1;
emuversion=src;
src+=strlen(src)+1;
description=src;
write_log("Saved with: '%s %s', description: '%s'\n",emuname,emuversion,description);
}

/* restore all subsystems */

void restore_state(char *filename)
{
FILE *f;
uae_u8 *chunk,*end;
char name[5];
int len;

chunk=0;
f=fopen(filename,"rb");
if(!f) goto error;

chunk=restore_chunk(f,name,&len);
if(!chunk||memcmp(name,"ASF ",4)) {
	write_log("%s is not an AmigaStateFile\n",filename);
	goto error;
}
restore_header(chunk);
free(chunk);
savestate_wanted=1;
for(;;) {
	chunk=restore_chunk(f,name,&len);
	if(!strcmp(name,"END ")) break;
	if(!strcmp(name,"CPU "))
		end=restore_cpu(chunk);
	else if(!strcmp(name,"CIAA"))
		end=restore_cia(0,chunk);
	else if(!strcmp(name,"CIAB"))
		end=restore_cia(1,chunk);
	else if(!strcmp(name,"CHIP"))
		end=restore_custom(chunk);
	else if(!strcmp(name,"DSK0"))
		end=restore_disk(0,chunk);
	else if(!strcmp(name,"DSK1"))
		end=restore_disk(1,chunk);
	else if(!strcmp(name,"DSK2"))
		end=restore_disk(2,chunk);
	else if(!strcmp(name,"DSK3"))
		end=restore_disk(3,chunk);
	else if(!strcmp(name,"RAM "))
		end=restore_ram(chunk);
	else if(!strcmp(name,"ROM "))
		end=restore_rom(chunk);
	if(len!=end-chunk) write_log("Chunk '%s' total size %d bytes but read %d bytes!\n",name,len,end-chunk);
	free(chunk);
}
error:
if(chunk) free(chunk);
if(f) fclose(f);
}

/* Save all subsystems  */

void save_state(char *filename, char *description)
{
uae_u8 header[1000];
uae_u8 *dst;
FILE *f;
int len,i;
char name[5];

f=fopen(filename,"wb");
if(!f) return;

dst=header;
save_long(0);
strcpy(dst,"UAE");
dst+=strlen(dst)+1;
/* Is there #define or variable that contains current version string? */
strcpy(dst,"0.8.7");
dst+=strlen(dst)+1;
strcpy(dst,description);
dst+=strlen(dst)+1;
save_chunk(f,header,dst-header,"ASF ");

dst=save_cpu(&len);
save_chunk(f,dst,len,"CPU ");
free(dst);

dst=save_custom(&len);
save_chunk(f,dst,len,"CHIP");
free(dst);

dst=save_cia(0,&len);
save_chunk(f,dst,len,"CIAA");
free(dst);

dst=save_cia(1,&len);
save_chunk(f,dst,len,"CIAB");
free(dst);

strcpy(name,"DSKx");
for(i=0;i<4;i++) {
	dst=save_disk(i,&len);
	if(dst) {
		name[3]=i+'0';
		save_chunk(f,dst,len,name);
		free(dst);
	}
}

dst=save_ram(1,&len);
do {
	save_chunk(f,dst,len,"RAM ");
	free(dst);
} while((dst=save_ram(0,&len)));

dst=save_rom(1,&len);
do {
	save_chunk(f,dst,len,"ROM ");
	free(dst);
} while((dst=save_rom(0,&len)));

fwrite("END ",1,4,f);
fwrite("\0\0\0\0",1,4,f);
write_log("Save of '%s' complete\n",filename);
fclose(f);
}

/*

My (Toni Wilen <twilen@sci.fi>) proposal
for Amiga-emulators' state-save format
Feel free to comment...
This is very similar to IFF-fileformat
Every hunk must end to 4 byte boundary,
fill with zero bytes if needed

version 0.4

HEADER

        hunk name ("ASF ") (AmigaStateFile)
        NOTE: Just noticed that extension "asf" ("Advanced Streaming Format")
              is registered with Media Player in Windows'98... this sucks.
        header size in bytes
        fileformat version number
        emulator name ("uae", "fellow" etc..)
        emulator version string (example: "0.8.7")
        free user writable comment string

CPU

        hunk name "CPU "
        hunk size in bytes

        CPU model               4 (68000,68010 etc..)
        CPU typeflags           bit 0=EC-model or not
        D0-D7                   8*4=32
        A0-A6                   7*4=32
        PC                      4
        prefetch                4
        USP                     4
        ISP                     4
        SR/CCR                  2
	flags			4 (bit 0=CPU was HALTed)

        CPU specific registers

        68000: SR/CCR is last saved register
        68010: save also DFC,SFC and VBR
        68020: all 68010 registers and CAAR,CACR and MSP
        etc..

        DFC                     4 (010+)
        SFC                     4 (010+)
        VBR                     4 (010+)

        CAAR                    4 (020-030)
        CACR                    4 (020+)
        MSP                     4 (020+)

MMU (when and if MMU is supported in future..)

FPU (only if used)

        hunk name "FPU "
        hunk size in bytes

        FPU model               4 (68881 or 68882)
        FPU typeflags           4 (keep zero)

        FP0-FP7                 4+2 (80 bits)
        FPCR                    4
        FPSR                    4
        FPIAR                   4

CUSTOM CHIPS

        hunk name "CHIP"
        hunk size in bytes

        chipset flags   4      OCS=0,ECSAGNUS=1,ECSDENISE=2,AGA=4
                               ECSAGNUS and ECSDENISE can be combined

        DFF000-DFF1FF   512

        this saves many unused or unneeded registers
        but is probably best solution

        AGA color               8 banks * 32 registers *
        registers               LONG (RGB) = 1024

        we should save these even if saving OCS/ECS state
        (better to have no difference between AGA and OCS/ECS fileformat)

        sprite "armed" status   8 sprites * 1 BYTE

        Sprites maybe armed in non-DMA mode
        Use bit 0 only, other bits are reserved

        internal audio state

        audio state             4 channels * 1 BYTE
        machine mode
        audio DMA               4 channels * 1 LONG
        pointers
        audio DMA
        length                  4 channels * 1 WORD
        audio period            4 channels * 1 WORD

        internal blitter state

        ????

CIA

        hunk name "CIAA" and "CIAB"
        hunk size in bytes (16+1+2*2+3+3+1=28)

        BFE001-BFEF01   16*1 (CIAA)
        BFD000-BFDF00   16*1 (CIAB)

        internal registers

        IRQ mask (ICR)  1 BYTE
        timer latches   2 timers * 2 BYTES (LO/HI)
        latched tod     3 BYTES (LO/MED/HI)
        alarm           3 BYTES (LO/MED/HI)
        flag            1 BYTE
                        bit 0=tod latched (read)
                        bit 1=tod stopped (write)

FLOPPY DRIVES

        hunk types:     "DSK0" "DSK1" "DSK2" "DSK3"
        hunk size in bytes

        drive state

        disk ID-word            2
        motor state             1 (off=0,on=1)
        rw-head track           1
        id-mode                 1 (-1=not in id-mode,>=0 number of bits
                                  transferred)
        floppy information

        words from              2
        beginning of track
        CRC of disk-image       4 (used during restore to check if image
                                  is correct)
        disk-image              null-terminated
        file name

        DMA state already saved with custom chip hunk
        (current disk DMA pointer and DMA length)
        and other floppy bits are saved with CIA hunks.

RAM SPACE

        hunk name "RAM "
        hunk size
        start of RAM "bank"     ("bank"=chip/slow/fast etc..)
        RAM "bank" size
        RAM flags               chip, slow or fast, need defining
        RAM "bank" contents

ROM SPACE

        hunk name "ROM "
        hunk size
        ROM start               4
        address
        size of ROM             4
        ROM flags               4 (not defined yet)
        ROM version             4
        ROM CRC                 4 see below
        ROM-image               null terminated, see below
        ID-string
        ROM contents            (Not mandatory, use hunk size to check if
                                this hunk contains ROM data or not)

        Kickstart ROM:
         ID-string is "Kickstart x.x"
         ROM version: version in high word and revision in low word
         Kickstart ROM version and revision can be found from ROM start
         + 12 (version) and +14 (revision)

        ROM version and CRC is only meant for emulator to automatically
        find correct image from its ROM-directory during state restore.

        Usually saving ROM contents is not good idea.


RAM/ROM alias (Is this needed?)

        hunk name "????"
        hunk size
        address where
        previous RAM/ROM
        "bank" also appears
        offset to RAM/ROM
        bank (usually zero)
        size of aliasing
        memory bank
        (usually same as
        last RAM/ROM bank
        size was)

        Use this hunk if last RAM/ROM bank appears more than one time in
        address space

END
        hunk name "END " ends


EMULATOR SPECIFIC HUNKS

Read only if "emulator name" in header is same as used emulator.

misc:

- save only at position 0,0
- blitter DMA:
   1. simple and may cause problems in some cases:
    force blitter-emulator to complete immediately
   2. complex:
    save and restore whole blitter-state. No, I am not going to do it :)
- disk DMA:
   1. simple
    if DMA is active, wait for next frame and try again etc..
   2. not so simple:
    save and restore state. No, I am not going to do this, either..
- saving other DMA states should be quite easy
  It seems only audio, disk and blitter DMA can be active in
  position 0,0
- all data probably should be saved in bigendian format
- should we strip all paths from image file names?
- I may have missed some "hidden" CIA and custom chip
  registers.
- note that I have not programmed hardware-banging Amiga programs
  for about 4 years... I probably have forgot many things.

*/